
/*
 * cb31main.c - Linux Host Driver for RATOC SCSI Adapters
 *
 * Copyright (c) 2001 RATOC Sysytems, Inc.
 * All Rights Reserved.
 *
 * Linux host driver for CardBus SCSI REX-CB31, CB31U.
 * We make this source from advansys.c
 *
 * Web Site:               http://www.ratocsystems.com
 * ----------------------------------------------------------
 *
 *  2001-10-31 v0.77 release
 *
 */

#define ASC_VERSION "3.3G"    /* AdvanSys Driver Version */

/*
 * advansys.c - Linux Host Driver for AdvanSys SCSI Adapters
 *
 * Copyright (c) 1995-2000 Advanced System Products, Inc.
 * Copyright (c) 2000-2001 ConnectCom Solutions, Inc.
 * All Rights Reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that redistributions of source
 * code retain the above copyright notice and this comment without
 * modification.
 *
 * As of March 8, 2000 Advanced System Products, Inc. (AdvanSys)
 * changed its name to ConnectCom Solutions, Inc.
 *
 * There is an AdvanSys Linux WWW page at:
 *  http://www.connectcom.net/downloads/software/os/linux.html
 *  http://www.advansys.com/linux.html
 *
 * The latest released version of the AdvanSys driver is available at:
 *  ftp://ftp.advansys.com/pub/linux/linux.tgz
 *  ftp://ftp.connectcom.net/pub/linux/linux.tgz
 *
 * Please send questions, comments, bug reports to:
 *  support@connectcom.net
 */

#define CVT_LINUX_VERSION(V, P, S)    (((V) * 65536) + ((P) * 256) + (S))

#ifndef LINUX_VERSION_CODE
#include <linux/version.h>
#endif

#ifdef MODULE
#include <linux/module.h>
#endif // MODULE

#if LINUX_VERSION_CODE < CVT_LINUX_VERSION(2,4,0)
#include <pcmcia/config.h>
#include <pcmcia/k_compat.h>
#endif // < 2.4.0

#include "cb31main.h"

#include <pcmcia/driver_ops.h>

#ifdef PCMCIA_DEBUG
static int pc_debug = PCMCIA_DEBUG;
MODULE_PARM(pc_debug, "i");
#define DEBUG(n, args...) if (pc_debug>(n)) printk(KERN_DEBUG args)
#else
#define DEBUG(n, args...)
#endif

static char *version =
"cb31_cb.c 0.75 2001/10/29 10:50:20";
/*===================================================================*/

/*
 * --- Linux 'Scsi_Host_Template' and cb31_setup() Functions
 */

/*===================================================================*/

typedef struct scsi_info_t {
    u_short	manf_id;
    int		ndev;
    dev_node_t	node[8];
} scsi_info_t;

int  init_module(void);
void cleanup_module(void);

#if LINUX_VERSION_CODE >= CVT_LINUX_VERSION(2,4,0)
static Scsi_Host_Template driver_template = CB31;
#else
#ifdef MODULE
Scsi_Host_Template driver_template = CB31;
#endif
#endif

#if LINUX_VERSION_CODE < CVT_LINUX_VERSION(2,3,27)
struct proc_dir_entry	proc_scsi_cb31 =
{
	PROC_SCSI_CB31, 4, "CB31",
	S_IFDIR | S_IRUGO | S_IXUGO, 2,
	0, 0, 0, NULL, NULL, NULL, NULL, NULL, NULL, NULL
};
#endif

static int ints[3] = { 0, 0, 0 };

void cb31_setup(char *, int *);

int cb31_Insert = 1;  /* 0: Not insert  1: Insert */
asc_board_t  *cb31_Insert_boardp = NULL;
static struct semaphore  sem_wait;  /* for host->eh_wait */

/*============================== for HotPlug =============================*/
#if CONFIG_HOTPLUG
static struct pci_device_id cb31_pci_table[] __devinitdata = {
	{ 0x10cd, 0x1300, 0x1195, 0x1320, },
	{ 0, },
};
MODULE_DEVICE_TABLE(pci, cb31_pci_table);

static int cb31_probe(struct pci_dev *pdev, const struct pci_device_id *id);
static void cb31_remove(struct pci_dev *pdev);

static struct pci_driver cb31_pci_ops = {
	name:		"cb31_cb",
	id_table:	cb31_pci_table,
	probe:		cb31_probe,
	remove:		cb31_remove,
	suspend:	NULL,
	resume:		NULL
};

struct Scsi_Host *cb31_dev_id;

#define PCI_POWERMGMT    0x40  // PCI register for PM

#else

static dev_node_t *cb31_attach(dev_locator_t *loc);
static void cb31_detach(dev_node_t *node);
static dev_node_t *cb31_config(dev_locator_t *loc);

struct driver_operations cb31_ops = {
    "cb31_cb", cb31_attach, NULL, NULL, cb31_detach
};


#endif



/*===================================================================*/

#if LINUX_VERSION_CODE >= CVT_LINUX_VERSION(1,3,0)
/*
 * cb31_proc_info() - /proc/scsi/advansys/[0-(ASC_NUM_BOARD_SUPPORTED-1)]
 *
 * *buffer: I/O buffer
 * **start: if inout == FALSE pointer into buffer where user read should start
 * offset: current offset into a /proc/scsi/advansys/[0...] file
 * length: length of buffer
 * hostno: Scsi_Host host_no
 * inout: TRUE - user is writing; FALSE - user is reading
 *
 * Return the number of bytes read from or written to a
 * /proc/scsi/advansys/[0...] file.
 *
 * Note: This function uses the per board buffer 'prtbuf' which is
 * allocated when the board is initialized in advansys_detect(). The
 * buffer is ASC_PRTBUF_SIZE bytes. The function asc_proc_copy() is
 * used to write to the buffer. The way asc_proc_copy() is written
 * if 'prtbuf' is too small it will not be overwritten. Instead the
 * user just won't get all the available statistics.
 */
int
cb31_proc_info(char *buffer, char **start, off_t offset, int length,
                   int hostno, int inout)
{
#ifdef CONFIG_PROC_FS
    struct Scsi_Host    *shp;
    asc_board_t         *boardp;
    int                 i;
    char                *cp;
    int                 cplen;
    int                 cnt;
    int                 totcnt;
    int                 leftlen;
    char                *curbuf;
    off_t               advoffset;
    Scsi_Device         *scd;

    ASC_DBG(1, "cb31_proc_info: begin\n");

    /*
     * User write not supported.
     */
    if (inout == TRUE) {
        return(-ENOSYS);
    }

    /*
     * User read of /proc/scsi/advansys/[0...] file.
     */

    /* Find the specified board. */
    for (i = 0; i < asc_board_count; i++) {
        if (asc_host[i]->host_no == hostno) {
            break;
        }
    }
    if (i == asc_board_count) {
        return(-ENOENT);
    }

    shp = asc_host[i];
    boardp = ASC_BOARDP(shp);

    /* Copy read data starting at the beginning of the buffer. */
    *start = buffer;
    curbuf = buffer;
    advoffset = 0;
    totcnt = 0;
    leftlen = length;

    /*
     * Get board configuration information.
     *
     * cb31_info() returns the board string from its own static buffer.
     */
    cp = (char *) cb31_info(shp);
    strcat(cp, "\n");
    cplen = strlen(cp);
    /* Copy board information. */
    cnt = asc_proc_copy(advoffset, offset, curbuf, leftlen, cp, cplen);
    totcnt += cnt;
    leftlen -= cnt;
    if (leftlen == 0) {
        ASC_DBG1(1, "cb31_proc_info: totcnt %d\n", totcnt);
        return totcnt;
    }
    advoffset += cplen;
    curbuf += cnt;

    /*
     * Display driver information for each device attached to the board.
     */
    cp = boardp->prtbuf;
    cplen = asc_prt_board_devices(shp, cp, ASC_PRTBUF_SIZE);
    ASC_ASSERT(cplen < ASC_PRTBUF_SIZE);
    cnt = asc_proc_copy(advoffset, offset, curbuf, leftlen, cp, cplen);
    totcnt += cnt;
    leftlen -= cnt;
    if (leftlen == 0) {
        ASC_DBG1(1, "cb31_proc_info: totcnt %d\n", totcnt);
        return totcnt;
    }
    advoffset += cplen;
    curbuf += cnt;

    /*
     * Display target driver information for each device attached
     * to the board.
     */
#if LINUX_VERSION_CODE < CVT_LINUX_VERSION(2,1,75)
    for (scd = scsi_devices; scd; scd = scd->next)
#else /* version >= v2.1.75 */
    for (scd = shp->host_queue; scd; scd = scd->next)
#endif /* version >= v2.1.75 */
    {
        if (scd->host == shp) {
            cp = boardp->prtbuf;
            /*
             * Note: If proc_print_scsidevice() writes more than
             * ASC_PRTBUF_SIZE bytes, it will overrun 'prtbuf'.
             */
            proc_print_scsidevice(scd, cp, &cplen, 0);
            ASC_ASSERT(cplen < ASC_PRTBUF_SIZE);
            cnt = asc_proc_copy(advoffset, offset, curbuf, leftlen, cp, cplen);
            totcnt += cnt;
            leftlen -= cnt;
            if (leftlen == 0) {
                ASC_DBG1(1, "advansys_proc_info: totcnt %d\n", totcnt);
                return totcnt;
            }
            advoffset += cplen;
            curbuf += cnt;
        }
    }

    /*
     * Display EEPROM configuration for the board.
     */
    cp = boardp->prtbuf;
    cplen = asc_prt_asc_board_eeprom(shp, cp, ASC_PRTBUF_SIZE);
    ASC_ASSERT(cplen < ASC_PRTBUF_SIZE);
    cnt = asc_proc_copy(advoffset, offset, curbuf, leftlen, cp, cplen);
    totcnt += cnt;
    leftlen -= cnt;
    if (leftlen == 0) {
        ASC_DBG1(1, "advansys_proc_info: totcnt %d\n", totcnt);
        return totcnt;
    }
    advoffset += cplen;
    curbuf += cnt;

    /*
     * Display driver configuration and information for the board.
     */
    cp = boardp->prtbuf;
    cplen = asc_prt_driver_conf(shp, cp, ASC_PRTBUF_SIZE);
    ASC_ASSERT(cplen < ASC_PRTBUF_SIZE);
    cnt = asc_proc_copy(advoffset, offset, curbuf, leftlen, cp, cplen);
    totcnt += cnt;
    leftlen -= cnt;
    if (leftlen == 0) {
        ASC_DBG1(1, "advansys_proc_info: totcnt %d\n", totcnt);
        return totcnt;
    }
    advoffset += cplen;
    curbuf += cnt;

#ifdef ADVANSYS_STATS
    /*
     * Display driver statistics for the board.
     */
    cp = boardp->prtbuf;
    cplen = asc_prt_board_stats(shp, cp, ASC_PRTBUF_SIZE);
    ASC_ASSERT(cplen < ASC_PRTBUF_SIZE);
    cnt = asc_proc_copy(advoffset, offset, curbuf, leftlen, cp, cplen);
    totcnt += cnt;
    leftlen -= cnt;
    if (leftlen == 0) {
        ASC_DBG1(1, "advansys_proc_info: totcnt %d\n", totcnt);
        return totcnt;
    }
    advoffset += cplen;
    curbuf += cnt;
#endif /* ADVANSYS_STATS */

    /*
     * Display Asc Library dynamic configuration information
     * for the board.
     */
    cp = boardp->prtbuf;
    cplen = asc_prt_asc_board_info(shp, cp, ASC_PRTBUF_SIZE);
    ASC_ASSERT(cplen < ASC_PRTBUF_SIZE);
    cnt = asc_proc_copy(advoffset, offset, curbuf, leftlen, cp, cplen);
    totcnt += cnt;
    leftlen -= cnt;
    if (leftlen == 0) {
        ASC_DBG1(1, "cb31_proc_info: totcnt %d\n", totcnt);
        return totcnt;
    }
    advoffset += cplen;
    curbuf += cnt;

    ASC_DBG1(1, "cb31_proc_info: totcnt %d\n", totcnt);

    return totcnt;
#else /* CONFIG_PROC_FS */
    return 0;
#endif /* CONFIG_PROC_FS */
} /* cb31_proc_info */
#endif /* version >= v1.3.0 */


/*
 * cb31_detect()
 *
 * Detect function for CB31 adapters.
 *
 * Argument is a pointer to the host driver's scsi_hosts entry.
 *
 * Return number of adapters found.
 *
 * Note: Because this function is called during system initialization
 * it must not call SCSI mid-level functions including scsi_malloc()
 * and scsi_free().
 */
int
cb31_detect(Scsi_Host_Template *tpnt)
{
    static int          detect_called = ASC_FALSE;
    int                 iop;
    struct Scsi_Host    *shp = NULL;
    asc_board_t         *boardp = NULL;
    ASC_DVC_VAR         *asc_dvc_varp = NULL;
    int                 ioport = 0;
    int                 share_irq = FALSE;
    int                 pci_init_search = 0;
    int                 pci_card_cnt_max = 0;
    int                 pci_card_cnt = 0;

    struct pci_dev      *pci_devicep[ASC_NUM_BOARD_SUPPORTED];
    struct pci_dev      *pci_devp = NULL;
    int                 pci_device_id_cnt = 0;
    unsigned int        pci_device_id[ASC_PCI_DEVICE_ID_CNT] = {
                                    ASC_PCI_DEVICE_ID_1100,
                                    ASC_PCI_DEVICE_ID_1200,
                                    ASC_PCI_DEVICE_ID_1300
                        };
    int                 warn_code, err_code;
    int                 ret;

   /*
    * Save the EEPROM configuration so that it can be displayed
    * from /proc/scsi/advansys/[0...].
    */
    ASCEEP_CONFIG *ep;


    if (detect_called == ASC_FALSE) {
        detect_called = ASC_TRUE;
    } else {
        printk("RATOC CardBus SCSI: cb31_detect() multiple calls ignored\n");
        return 0;
    }

//    printk("cb31_detect: begin\n");
    /*
     * XXX - Remove this comment and the next line when SCSI mid-level
     * no longer acquires 'io_request_lock' before calling the SCSI
     * low-level detect entrypoint.
     */
    ASC_UNLOCK_IO_REQUEST_LOCK

#if LINUX_VERSION_CODE < CVT_LINUX_VERSION(2,3,27) // 2001-05-31
    tpnt->proc_dir = &proc_scsi_cb31;
#else
    tpnt->proc_name = "CB31";
#endif  /* version <= v2.3.27 */

    asc_board_count = 0;

    /*
     * If I/O port probing has been modified, then verify and
     * clean-up the 'asc_ioport' list.
     */
    if (asc_iopflag == ASC_TRUE) {
        for (ioport = 0; ioport < ASC_NUM_IOPORT_PROBE; ioport++) {
            if (asc_ioport[ioport] != 0) {
                for (iop = 0; iop < ASC_IOADR_TABLE_MAX_IX; iop++) {
                    if (_asc_def_iop_base[iop] == asc_ioport[ioport]) {
                        break;
                    }
                } // for
                if (iop == ASC_IOADR_TABLE_MAX_IX) {
                    printk(
"cb31_detect: specified I/O Port 0x%X is invalid\n",
                        asc_ioport[ioport]);
                    asc_ioport[ioport] = 0;
                }
            }
        } // for (ioport...)
        ioport = 0;
    } // if (asc_iopflag ...)

    iop = 0;

    while (asc_board_count < ASC_NUM_BOARD_SUPPORTED) {

        ASC_DBG1(2, "cb31_detect: asc_board_count %d\n", asc_board_count);

        if (pci_init_search == 0) {
            int i, j;

            pci_init_search = 1;

            /* Find all PCI cards. */
            while (pci_device_id_cnt < ASC_PCI_DEVICE_ID_CNT) {
                if ((pci_devp = pci_find_device(ASC_PCI_VENDORID,
                    pci_device_id[pci_device_id_cnt], pci_devp)) ==
                    NULL) {
                    pci_device_id_cnt++;
                } else {
                    pci_devicep[pci_card_cnt_max++] = pci_devp;
                }
            }  // while

            /*
             * Sort PCI cards in ascending order by PCI Bus, Slot,
             * and Device Number.
             */
            for (i = 0; i < pci_card_cnt_max - 1; i++) {
                for (j = i + 1; j < pci_card_cnt_max; j++) {
                    if ((pci_devicep[j]->bus->number < pci_devicep[i]->bus->number) ||
                        ((pci_devicep[j]->bus->number == pci_devicep[i]->bus->number) &&
                          (pci_devicep[j]->devfn < pci_devicep[i]->devfn))) {
                            pci_devp = pci_devicep[i];
                            pci_devicep[i] = pci_devicep[j];
                            pci_devicep[j] = pci_devp;
                    } // if
                }  // for (j...)
            } // for (i...)

            pci_card_cnt = 0;
        } else {
            pci_card_cnt++;
        } // if (pci_init_search...)

        if (pci_card_cnt == pci_card_cnt_max) {
            iop = 0;
        } else {
            pci_devp = pci_devicep[pci_card_cnt];

#if LINUX_VERSION_CODE < CVT_LINUX_VERSION(2,3,13)
            iop = pci_devp->base_address[0] & PCI_IOADDRESS_MASK;
#else    /* version >= v2.3.13 */
//            iop = pci_devp->resource[1].start & PCI_IOADDRESS_MASK;
            iop = pci_resource_start(pci_devp, 0);
#endif   /* version >= v2.3.13 */
            ASC_DBG2(1,
                "cb31_detect: vendorID %X, deviceID %X\n",
                pci_devp->vendor, pci_devp->device);
            ASC_DBG2(2, "cb31_detect: iop %X, irqLine %d\n",
                iop, pci_devp->irq);
        }

        /*
         * Adapter not found, try next bus type.
         */
        if (iop == 0) {
            break;
        }

        /*
         * Adapter found.
         *
         * Register the adapter, get its configuration, and
         * initialize it.
         */
        shp = scsi_register(tpnt, sizeof(asc_board_t));
        if (shp == NULL) {
            printk("scsi_register() error\n");
            break;
        }

#if LINUX_VERSION_CODE >= CVT_LINUX_VERSION(2,4,4)
        scsi_set_pci_device(shp, pci_devp);  /* 2001-10-25 */
#endif

        /* Save a pointer to the Scsi_host of each board found. */
        asc_host[asc_board_count++] = shp;

        /* Initialize private per board data */
        boardp = ASC_BOARDP(shp);
        memset(boardp, 0, sizeof(asc_board_t));
        boardp->id = asc_board_count - 1;

        /* Initialize spinlock. */
        boardp->lock = SPIN_LOCK_UNLOCKED;

        /*
         * Handle narrow boards.
         */

        asc_dvc_varp = &boardp->dvc_var.asc_dvc_var;
        asc_dvc_varp->bus_type = ASC_IS_PCI;
        asc_dvc_varp->drv_ptr = (ulong) boardp;
        asc_dvc_varp->cfg = &boardp->dvc_cfg.asc_dvc_cfg;
        asc_dvc_varp->cfg->overrun_buf = &overrun_buf[0];
        asc_dvc_varp->iop_base = iop;
        asc_dvc_varp->isr_callback = asc_isr_callback;
//        asc_dvc_varp->isr_callback = (Ptr2Func) asc_isr_callback;

        /*
         * Allocate buffer for printing information from
         * /proc/scsi/advansys/[0...].
         */
        if ((boardp->prtbuf =
            kmalloc(ASC_PRTBUF_SIZE, GFP_ATOMIC)) == NULL) {
            ASC_PRINT3(
"cb31_detect: board %d: kmalloc(%d, %d) returned NULL\n",
                boardp->id, ASC_PRTBUF_SIZE, GFP_ATOMIC);
            scsi_unregister(shp);
            asc_board_count--;
            continue;
        }


        /*
         * Set the board bus type and PCI IRQ before
         * calling AscInitGetConfig().
         */
        shp->irq = asc_dvc_varp->irq_no = pci_devp->irq;
        asc_dvc_varp->cfg->pci_device_id = pci_devp->device;
        asc_dvc_varp->cfg->pci_slot_info =
            ASC_PCI_MKID(pci_devp->bus->number,
                PCI_SLOT(pci_devp->devfn),
                PCI_FUNC(pci_devp->devfn));
        shp->unchecked_isa_dma = FALSE;
        share_irq = TRUE;

        /*
         * Read the board configuration.
         *
         * NOTE: AscInitGetConfig() may change the board's
         * bus_type value. The asc_bus[bus] value should no
         * longer be used. If the bus_type field must be
         * referenced only use the bit-wise AND operator "&".
         */
        switch(ret = AscInitGetConfig(asc_dvc_varp)) {
        case 0:    /* No error */
            break;
        case ASC_WARN_IO_PORT_ROTATE:
            ASC_PRINT1(
"AscInitGetConfig: board %d: I/O port address modified\n",
                boardp->id);
            break;
        case ASC_WARN_AUTO_CONFIG:
            ASC_PRINT1(
"AscInitGetConfig: board %d: I/O port increment switch enabled\n",
                boardp->id);
            break;
        case ASC_WARN_EEPROM_CHKSUM:
            ASC_PRINT1(
"AscInitGetConfig: board %d: EEPROM checksum error\n",
                boardp->id);
            break;
        case ASC_WARN_IRQ_MODIFIED:
            ASC_PRINT1(
"AscInitGetConfig: board %d: IRQ modified\n",
                boardp->id);
            break;
        case ASC_WARN_CMD_QNG_CONFLICT:
            ASC_PRINT1(
"AscInitGetConfig: board %d: tag queuing enabled w/o disconnects\n",
                boardp->id);
            break;
        default:
            ASC_PRINT2(
"AscInitGetConfig: board %d: unknown warning: %x\n",
                boardp->id, ret);
            break;
        } // switch(ret = AscInitGetConfig(asc_dvc_varp))
        if ((err_code = asc_dvc_varp->err_code) != 0) {
            ASC_PRINT3(
"AscInitGetConfig: board %d error: init_state %x, err_code %x\n",
                boardp->id, asc_dvc_varp->init_state,
                asc_dvc_varp->err_code);
        } // if (err_code...)

        if (err_code != 0) {
#ifdef CONFIG_PROC_FS
            kfree(boardp->prtbuf);
#endif /* CONFIG_PROC_FS */
            scsi_unregister(shp);
            asc_board_count--;
            continue;
        } // if (err_code...)


        /*
         * Set the adapter's target id bit in the 'init_tidmask' field.
         */
        boardp->init_tidmask |=
            ADV_TID_TO_TIDMASK(asc_dvc_varp->cfg->chip_scsi_id);

        /*
         * Save EEPROM settings for the board.
         */
        ep = &boardp->eep_config.asc_eep;

        ep->init_sdtr = asc_dvc_varp->cfg->sdtr_enable;
        ep->disc_enable = asc_dvc_varp->cfg->disc_enable;
        ep->use_cmd_qng = asc_dvc_varp->cfg->cmd_qng_enabled;
        ep->isa_dma_speed = asc_dvc_varp->cfg->isa_dma_speed;
        ep->start_motor = asc_dvc_varp->start_motor;
        ep->cntl = asc_dvc_varp->dvc_cntl;
        ep->no_scam = asc_dvc_varp->no_scam;
        ep->max_total_qng = asc_dvc_varp->max_total_qng;
        ep->chip_scsi_id = asc_dvc_varp->cfg->chip_scsi_id;
        /* 'max_tag_qng' is set to the same value for every device. */
        ep->max_tag_qng = asc_dvc_varp->cfg->max_tag_qng[0];
        ep->adapter_info[0] = asc_dvc_varp->cfg->adapter_info[0];
        ep->adapter_info[1] = asc_dvc_varp->cfg->adapter_info[1];
        ep->adapter_info[2] = asc_dvc_varp->cfg->adapter_info[2];
        ep->adapter_info[3] = asc_dvc_varp->cfg->adapter_info[3];
        ep->adapter_info[4] = asc_dvc_varp->cfg->adapter_info[4];
        ep->adapter_info[5] = asc_dvc_varp->cfg->adapter_info[5];
        ep->adapter_info[6] = asc_dvc_varp->cfg->adapter_info[6];

       /*
        * Modify board configuration.
        */
        switch (ret = AscInitSetConfig(asc_dvc_varp)) {
        case 0:    /* No error. */
            break;
        case ASC_WARN_IO_PORT_ROTATE:
            ASC_PRINT1(
"AscInitSetConfig: board %d: I/O port address modified\n",
                boardp->id);
            break;
        case ASC_WARN_AUTO_CONFIG:
            ASC_PRINT1(
"AscInitSetConfig: board %d: I/O port increment switch enabled\n",
                boardp->id);
            break;
        case ASC_WARN_EEPROM_CHKSUM:
            ASC_PRINT1(
"AscInitSetConfig: board %d: EEPROM checksum error\n",
                boardp->id);
            break;
        case ASC_WARN_IRQ_MODIFIED:
            ASC_PRINT1(
"AscInitSetConfig: board %d: IRQ modified\n",
                    boardp->id);
            break;
        case ASC_WARN_CMD_QNG_CONFLICT:
            ASC_PRINT1(
"AscInitSetConfig: board %d: tag queuing w/o disconnects\n",
                    boardp->id);
            break;
        default:
            ASC_PRINT2(
"AscInitSetConfig: board %d: unknown warning: %x\n",
                    boardp->id, ret);
            break;
        } // switch (ret = AscInitSetConfig(asc_dvc_varp))
        if (asc_dvc_varp->err_code != 0) {
            ASC_PRINT3(
"AscInitSetConfig: board %d error: init_state %x, err_code %x\n",
                boardp->id, asc_dvc_varp->init_state,
                asc_dvc_varp->err_code);
#ifdef CONFIG_PROC_FS
            kfree(boardp->prtbuf);
#endif /* CONFIG_PROC_FS */
            scsi_unregister(shp);
            asc_board_count--;
            continue;
        }

        /*
         * Finish initializing the 'Scsi_Host' structure.
         */

        /*
         * Channels are numbered beginning with 0. For AdvanSys One host
         * structure supports one channel. Multi-channel boards have a
         * separate host structure for each channel.
         */
        shp->max_channel = 0;
        shp->max_id = ASC_MAX_TID + 1;
        shp->max_lun = ASC_MAX_LUN + 1;

        shp->io_port = asc_dvc_varp->iop_base;
        shp->n_io_port = ASC_IOADR_GAP;
        shp->this_id = asc_dvc_varp->cfg->chip_scsi_id;

        /* Set maximum number of queues the adapter can handle. */
        shp->can_queue = asc_dvc_varp->max_total_qng;

        /*
         * 'n_io_port' currently is one byte.
         *
         * Set a value to 'n_io_port', but never referenced it because
         * it may be truncated.
         */
        shp->n_io_port = boardp->asc_n_io_port <= 255 ?
            boardp->asc_n_io_port : 255;

        /*
         * Following v1.3.89, 'cmd_per_lun' is no longer needed
         * and should be set to zero.
         *
         * But because of a bug introduced in v1.3.89 if the driver is
         * compiled as a module and 'cmd_per_lun' is zero, the Mid-Level
         * SCSI function 'allocate_device' will panic. To allow the driver
         * to work as a module in these kernels set 'cmd_per_lun' to 1.
         */
#ifdef MODULE
        shp->cmd_per_lun = 1;
#else   /* MODULE */
        shp->cmd_per_lun = 0;
#endif  /* MODULE */
        /*
         * Use the host 'select_queue_depths' function to determine
         * the number of commands to queue per device.
         */
        shp->select_queue_depths = cb31_select_queue_depths;

        /*
         * Set the maximum number of scatter-gather elements the
         * adapter can handle.
         */
        /*
         * Allow two commands with 'sg_tablesize' scatter-gather
         * elements to be executed simultaneously. This value is
         * the theoretical hardware limit. It may be decreased
         * below.
         */
        shp->sg_tablesize =
            (((asc_dvc_varp->max_total_qng - 2) / 2) *
            ASC_SG_LIST_PER_Q) + 1;

#ifdef MODULE
        /*
         * If the driver is compiled as a module, set a limit on the
         * 'sg_tablesize' value to prevent memory allocation failures.
         * Memory allocation errors are more likely to occur at module
         * load time, then at driver initialization time.
         */
        if (shp->sg_tablesize > 64) {
            shp->sg_tablesize = 64;
        }
#endif /* MODULE */

        /*
         * The value of 'sg_tablesize' can not exceed the SCSI
         * mid-level driver definition of SG_ALL. SG_ALL also
         * must not be exceeded, because it is used to define the
         * size of the scatter-gather table in 'struct asc_sg_head'.
         */
        if (shp->sg_tablesize > SG_ALL) {
            shp->sg_tablesize = SG_ALL;
        }

        ASC_DBG1(1, "cb31_detect: sg_tablesize: %d\n", shp->sg_tablesize);

        /* BIOS start address. */
        shp->base = 0; /* No BIOS */

        /*
         * Register Board Resources - I/O Port, DMA, IRQ
         */

        /* Register I/O port range. */
        request_region(shp->io_port, shp->n_io_port, "cb31");

        /* Register DMA Channel for Narrow boards. */
        shp->dma_channel = NO_ISA_DMA; /* Default to no ISA DMA. */

        /* Register IRQ Number. */

       /*
        * If request_irq() fails with the SA_INTERRUPT flag set,
        * then try again without the SA_INTERRUPT flag set. This
        * allows IRQ sharing to work even with other drivers that
        * do not set the SA_INTERRUPT flag.
        *
        * If SA_INTERRUPT is not set, then interrupts are enabled
        * before the driver interrupt function is called.
        */
        if (((ret = request_irq(shp->irq, cb31_interrupt,
                        SA_INTERRUPT | (share_irq == TRUE ? SA_SHIRQ : 0),
                        "cb31", boardp)) != 0) &&
            ((ret = request_irq(shp->irq, cb31_interrupt,
                        (share_irq == TRUE ? SA_SHIRQ : 0),
                        "cb31", boardp)) != 0))
        {
            if (ret == -EBUSY) {
                ASC_PRINT2(
"cb31_detect: board %d: request_irq(): IRQ %d already in use.\n",
                    boardp->id, shp->irq);
            } else if (ret == -EINVAL) {
                ASC_PRINT2(
"cb31_detect: board %d: request_irq(): IRQ %d not valid.\n",
                    boardp->id, shp->irq);
            } else {
                ASC_PRINT3(
"cb31_detect: board %d: request_irq(): IRQ %d failed with %d\n",
                    boardp->id, shp->irq, ret);
            }
            release_region(shp->io_port, shp->n_io_port);
            iounmap(boardp->ioremap_addr);
            if (shp->dma_channel != NO_ISA_DMA) {
                free_dma(shp->dma_channel);
            }
#ifdef CONFIG_PROC_FS
            kfree(boardp->prtbuf);
#endif /* CONFIG_PROC_FS */
            scsi_unregister(shp);
            asc_board_count--;
            continue;
        } // if ((ret = request_irq

        /*
         * Initialize board RISC chip and enable interrupts.
         */
        warn_code = AscInitAsc1000Driver(asc_dvc_varp);
        err_code = asc_dvc_varp->err_code;

        if (warn_code || err_code) {
            ASC_PRINT4(
"AscInitAsc1000Driver: board %d: error: init_state %x, warn %x error %x\n",
                boardp->id, asc_dvc_varp->init_state,
                warn_code, err_code);
        }

        if (err_code != 0) {
            release_region(shp->io_port, shp->n_io_port);
            if (shp->dma_channel != NO_ISA_DMA) {
                free_dma(shp->dma_channel);
            }
#ifdef CONFIG_PROC_FS
            kfree(boardp->prtbuf);
#endif /* CONFIG_PROC_FS */
            free_irq(shp->irq, boardp);
            scsi_unregister(shp);
            asc_board_count--;
            continue;
        } // if (err_code ...)

        ASC_DBG_PRT_SCSI_HOST(2, shp);
    } // while (asc_board_count...)
    printk("cb31_cb: Total Adapters = %d\n", asc_board_count);
    ASC_DBG1(1, "cb31_detect: done: asc_board_count %d\n", asc_board_count);
    return asc_board_count;
} /* cb31_detect */

/*
 * cb31_release()
 *
 * Release resources allocated for a single CB31 adapter.
 */
int
cb31_release(struct Scsi_Host *shp)
{
    asc_board_t    *boardp;

//  printk("cb31_release: begin\n");

    ASC_DBG(1, "cb31_release: begin\n");
    boardp = ASC_BOARDP(shp);
//  printk("cb31_release: AscStopChip\n");
    AscStopChip(shp->io_port); /* Stop SCSI Controler */
    free_irq(shp->irq, boardp);
    release_region(shp->io_port, shp->n_io_port);
#ifdef CONFIG_PROC_FS
    ASC_ASSERT(boardp->prtbuf != NULL);
    kfree(boardp->prtbuf);
#endif /* CONFIG_PROC_FS */
    scsi_unregister(shp);
    ASC_DBG(1, "cb31_release: end\n");
    return 0;
} /* cb31_release */

/*
 * cb31_info()
 *
 * Return suitable for printing on the console with the argument
 * adapter's configuration information.
 *
 * Note: The information line should not exceed ASC_INFO_SIZE bytes,
 * otherwise the static 'info' array will be overrun.
 */
const char *
cb31_info(struct Scsi_Host *shp)
{
    static char     info[ASC_INFO_SIZE];
    asc_board_t     *boardp;
    ASC_DVC_VAR     *asc_dvc_varp;

    boardp = ASC_BOARDP(shp);
    asc_dvc_varp = &boardp->dvc_var.asc_dvc_var;
    ASC_DBG(1, "cb31_info: begin\n");
    sprintf(info, "RATOC UltraSCSI CB31 %s: %s %u CDB: IO %lX/%X, IRQ %u",
            ASC_VERSION, "CardBus", asc_dvc_varp->max_total_qng,
            shp->io_port, shp->n_io_port - 1, shp->irq);

    ASC_ASSERT(strlen(info) < ASC_INFO_SIZE);
    ASC_DBG(1, "cb31_info: end\n");
    return info;
} /* cb31_info */

/*
 * cb31_command() - polled I/O entrypoint.
 *
 * Apparently host drivers shouldn't return until the command
 * is finished.
 *
 * Note: This is an old interface that is no longer used by the SCSI
 * mid-level driver. The new interface, advansys_queuecommand(),
 * currently handles all requests.
 */
int
cb31_command(Scsi_Cmnd *scp)
{
    ASC_DBG1(1, "cb31_command: scp %x\n", (unsigned) scp);
    ASC_STATS(scp->host, command);
    scp->SCp.Status = 0; /* Set to a known state */
    cb31_queuecommand(scp, cb31_command_done);
    while (scp->SCp.Status == 0) {
        /* 2001-10-22 */
        if (cb31_Insert == 0) {
            printk("cb31_command: card is not inserted\n");
            break;
        }
        /* 2001-10-22 */
        continue;
    }
    ASC_DBG1(1, "cb31_command: result %x\n", scp->result);
    return scp->result;
} /* cb31_command */

/*
 * cb31_queuecommand() - interrupt-driven I/O entrypoint.
 *
 * This function always returns 0. Command return status is saved
 * in the 'scp' result field.
 */

int
cb31_queuecommand(Scsi_Cmnd *scp, void (*done)(Scsi_Cmnd *))
{
    struct Scsi_Host    *shp;
    asc_board_t         *boardp;
    ulong               flags;
    Scsi_Cmnd           *done_scp;

//  printk("cb31_queuecommand: start\n");
    /* 2001-10-22 */
//    if (cb31_Insert == 0) {
//        printk("cb31_queuecommand: card is not inserted\n");
//        return ERR;
//    }
    /* 2001-10-22 */

    shp = scp->host;
    boardp = ASC_BOARDP(shp);
    ASC_STATS(shp, queuecommand);

    /*
     * XXX - Remove this comment and the next line when SCSI mid-level
     * no longer acquires 'io_request_lock' before calling the SCSI
     * low-level queuecommand entrypoint.
     */

    ASC_UNLOCK_IO_REQUEST_LOCK

    spin_lock_irqsave(&boardp->lock, flags);

    /*
     * Block new commands while handling a reset or abort request.
     */
    if (boardp->flags & ASC_HOST_IN_RESET) {
        ASC_DBG1(1,
            "cb31_queuecommand: scp 0x%lx blocked for reset request\n",
            (ulong) scp);
        scp->result = HOST_BYTE(DID_RESET);

    /* 2001-10-27 test */
    if (cb31_Insert == 0) {
        printk("cb31_queuecommand: card is not inserted\n");
//        done_scp = asc_dequeue_list(&boardp->done, NULL, ASC_TID_ALL);
        /* Interrupts could be enabled here. */
//        asc_scsi_done_list(done_scp);
        spin_unlock_irqrestore(&boardp->lock, flags);
        return ERR;
    }
    /* 2001-10-27 test */
        /*
         * Add blocked requests to the board's 'done' queue. The queued
         * requests will be completed at the end of the abort or reset
         * handling.
         */
        asc_enqueue(&boardp->done, scp, ASC_BACK);
        spin_unlock_irqrestore(&boardp->lock, flags);
        return 0;
    }

    /*
     * Attempt to execute any waiting commands for the board.
     */
    if (!ASC_QUEUE_EMPTY(&boardp->waiting)) {
        ASC_DBG(1,
            "cb31_queuecommand: before asc_execute_queue() waiting\n");
        asc_execute_queue(&boardp->waiting);
    }

    /*
     * Save the function pointer to Linux mid-level 'done' function
     * and attempt to execute the command.
     *
     * If ASC_NOERROR is returned the request has been added to the
     * board's 'active' queue and will be completed by the interrupt
     * handler.
     *
     * If ASC_BUSY is returned add the request to the board's per
     * target waiting list. This is the first time the request has
     * been tried. Add it to the back of the waiting list. It will be
     * retried later.
     *
     * If an error occurred, the request will have been placed on the
     * board's 'done' queue and must be completed before returning.
     */
    scp->scsi_done = done;
    switch (asc_execute_scsi_cmnd(scp)) {
    case ASC_NOERROR:
        break;
    case ASC_BUSY:
        asc_enqueue(&boardp->waiting, scp, ASC_BACK);
        break;
    case ASC_ERROR:
    default:
        done_scp = asc_dequeue_list(&boardp->done, NULL, ASC_TID_ALL);
        /* Interrupts could be enabled here. */
        asc_scsi_done_list(done_scp);
        break;
    }

    spin_unlock_irqrestore(&boardp->lock, flags);

    /*
     * XXX - Remove this comment and the next line when SCSI mid-level
     * no longer acquires 'io_request_lock' before calling the SCSI
     * low-level queuecommand entrypoint.
     */
    ASC_LOCK_IO_REQUEST_LOCK

    return 0;
}  /* cb31_queuecommand */


/*
 * cb31_abort()
 *
 * Abort the command specified by 'scp'.
 */
int
cb31_abort(Scsi_Cmnd *scp)
{
    struct Scsi_Host    *shp;
    asc_board_t         *boardp;
    ASC_DVC_VAR         *asc_dvc_varp;
    int                 flags;
    int                 do_scsi_done;
    int                 scp_found;
    Scsi_Cmnd           *done_scp = NULL;
    int                 ret;
    int                 chipstat;

//  printk("cb31_abort : start\n");
    /* Save current flags and disable interrupts. */
    save_flags(flags);
    cli();

    ASC_DBG1(1, "cb31_abort: scp %x\n", (unsigned) scp);

#ifdef ADVANSYS_STATS
    if (scp->host != NULL) {
        ASC_STATS(scp->host, abort);
    }
#endif /* ADVANSYS_STATS */


    if (scp->serial_number != scp->serial_number_at_timeout) {
        ASC_PRINT1(
"cb31_abort: timeout serial number changed for request %x\n",
            (unsigned) scp);
  DvcSleepMilliSecond(2000L);
        do_scsi_done = ASC_FALSE;
        scp_found = ASC_FALSE;
        ret = SCSI_ABORT_NOT_RUNNING;
    }
    else /* if (scp->serial_number...) */
    if (cb31_Insert == 0) {
        printk("cb31_abort: card is removed\n");
        scp->result = HOST_BYTE(DID_NO_CONNECT);
        scp->eh_state = FAILED;
        scp->host->in_recovery = 1;
        scp->device->online = FALSE;
        ret = SCSI_ABORT_ERROR;
        restore_flags(flags);
        return(ret);
    } /* if (b31_Insert... */
    else
    if ((shp = scp->host) == NULL) {
        scp->result = HOST_BYTE(DID_ERROR);
        do_scsi_done = ASC_TRUE;
        scp_found = ASC_FALSE;
        ret = SCSI_ABORT_ERROR;
    } else if ((boardp = ASC_BOARDP(shp))->flags &
                (ASC_HOST_IN_RESET | ASC_HOST_IN_ABORT)) {
        ASC_PRINT2(
"cb31_abort: board %d: Nested host reset or abort, flags 0x%x\n",
            boardp->id, boardp->flags);
        do_scsi_done = ASC_TRUE;
        if ((asc_rmqueue(&boardp->active, scp) == ASC_TRUE) ||
            (asc_rmqueue(&boardp->waiting, scp) == ASC_TRUE)) {
            scp_found = ASC_TRUE;
        } else {
            scp_found = ASC_FALSE;
        }
        scp->result = HOST_BYTE(DID_ERROR);
        ret = SCSI_ABORT_ERROR;
    } else {  // if ((boardp = ASC_BOARDP(shp))->flags ...

        /* Set abort flag to avoid nested reset or abort requests. */
        boardp->flags |= ASC_HOST_IN_ABORT;

        do_scsi_done = ASC_TRUE;
        if (asc_rmqueue(&boardp->waiting, scp) == ASC_TRUE) {
            /*
             * If asc_rmqueue() found the command on the waiting
             * queue, it had not been sent to the device. After
             * the queue is removed, no other handling is required.
             */
            ASC_PRINT1("cb31_abort: scp %x found on waiting queue\n",
                (unsigned) scp);
            ASC_DBG1(1, "cb31_abort: scp %x found on waiting queue\n",
                (unsigned) scp);
            scp_found = ASC_TRUE;
//            scp->result = HOST_BYTE(DID_ABORT);
//            ret = SCSI_ABORT_SUCCESS;
        scp->result = HOST_BYTE(DID_ERROR); /* 2001-10-18 test */
        ret = SCSI_ABORT_ERROR; /* 2001-10-18 test */
        } else if (asc_isqueued(&boardp->active, scp) == ASC_TRUE) {
            /*
             * If asc_isqueued() found the command on the active
             * queue, it has been sent to the device. The command
             * will be returned through the interrupt handler after
             * it has been aborted.
             */

            asc_dvc_varp = &boardp->dvc_var.asc_dvc_var;
            scp->result = HOST_BYTE(DID_ABORT);

            /* sti(); XXX */ /* Enable interrupts for AscAbortSRB(). */
            /* printk("cb31_abort: before AscAbortSRB(), scp %x\n",
                    (unsigned) scp);  */
            switch (AscAbortSRB(asc_dvc_varp, (ulong) scp)) {
            case ASC_TRUE:
                /* asc_isr_callback() will be called */
            //  printk("cb31_abort: AscAbortSRB() TRUE\n");
                ret = SCSI_ABORT_PENDING;
                break;
            case ASC_FALSE:
                /* Request has apparently already completed. */
            //  printk("cb31_abort: AscAbortSRB() FALSE\n");
                ret = SCSI_ABORT_NOT_RUNNING;
                break;
            case ASC_ERROR:
            default:
            //  printk("cb31_abort: AscAbortSRB() ERROR\n");
                ret = SCSI_ABORT_ERROR;
                break;
            } // switch (AscAbortSRB()...)
            cli();

            /*
             * The request will either still be on the active queue
             * or have been added to the board's done queue.
             */
            if (asc_rmqueue(&boardp->active, scp) == ASC_TRUE) {
                scp->result = HOST_BYTE(DID_ABORT);
                scp_found = ASC_TRUE;
        // printk("asc_rmqueue() == ASC_TRUE\n");
            } else {
                scp_found = asc_rmqueue(&boardp->done, scp);
                ASC_ASSERT(scp_found == ASC_TRUE);
//        printk("asc_rmqueue() != ASC_TRUE\n"); /* 2001-10-18 test */
          scp->result = HOST_BYTE(DID_ERROR); /* 2001-10-18 test */
          scp->eh_state = FAILED; /* 2001-10-19 test */
          ret = SCSI_ABORT_ERROR; /* 2001-10-18 test */
            }
        } else { // if (asc_isqueued() == ASC_TRUE )
            /*
             * The command was not found on the active or waiting queues.
             */
            do_scsi_done = ASC_TRUE;
            scp_found = ASC_FALSE;
            ret = SCSI_ABORT_NOT_RUNNING;
           //  printk("cb31_abort: NOT RUNNING\n");
        } // if (asc_isqueued() == ASC_TRUE )

        /* Clear abort flag. */
        boardp->flags &= ~ASC_HOST_IN_ABORT;

        /*
         * Because the ASC_HOST_IN_ABORT flag causes both
         * 'cb31_interrupt' and 'asc_isr_callback' to
         * queue requests to the board's 'done' queue and
         * prevents waiting commands from being executed,
         * these queued requests must be handled here.
         */
        done_scp = asc_dequeue_list(&boardp->done, NULL, ASC_TID_ALL);

        /*
         * Start any waiting commands for the board.
         */
        if (!ASC_QUEUE_EMPTY(&boardp->waiting)) {
          chipstat = AscGetChipStatus(boardp->dvc_var.asc_dvc_var.iop_base);
          if (chipstat & CSW_PCIERR) {
            printk("cb31_abort: detect PCIERR.\n");
            scp->result = HOST_BYTE(DID_ERROR);
            scp->host->in_recovery = 1;
            if (scp->host->eh_wait == NULL)
                scp->host->eh_wait = &sem_wait;
            scp->eh_state = FAILED;
            scp->device->online = FALSE;
            if (done_scp != NULL) {
                done_scp->result = scp->result;
                done_scp->eh_state = scp->eh_state;
                done_scp->device = scp->device;
                done_scp->host->eh_wait = scp->host->eh_wait;
            }
            ret = SCSI_ABORT_ERROR;
          } else {
//     printk("cb31_abort: asc_execute_queue()\n");
            ASC_DBG(1, "cb31_abort: before asc_execute_queue()\n");
            asc_execute_queue(&boardp->waiting);
          }
        }
    } // if ((boardp = ASC_BOARDP(shp))->flags ...

    /* Interrupts could be enabled here. */

    /*
     * Complete the request to be aborted, unless it has been
     * restarted as detected above, even if it was not found on
     * the device active or waiting queues.
     */
    ASC_ASSERT(do_scsi_done != ASC_ERROR);
    ASC_ASSERT(scp_found != ASC_ERROR);
    if (do_scsi_done == ASC_TRUE) {
        if (scp->scsi_done == NULL) {
            ASC_PRINT1(
"cb31_abort: aborted request scsi_done() is NULL, %x\n",
                (unsigned) scp);
        } else {
            if (scp_found == ASC_FALSE) {
                ASC_PRINT1(
"cb31_abort: abort request not active or waiting, completing anyway %x\n",
                    (unsigned) scp);
            }
            ASC_STATS(scp->host, done);
            scp->scsi_done(scp);
        }
    } // if (do_scsi_done == ASC_TRUE)

    /*
     * It is possible for the request done function to re-enable
     * interrupts without confusing the driver. But here interrupts
     * aren't enabled until all requests have been completed.
     */
    if (done_scp != NULL) {
        asc_scsi_done_list(done_scp);
    }

    ASC_DBG1(1, "cb31_abort: ret %d\n", ret);

    /* Re-enable interrupts, if they were enabled on entry. */
    restore_flags(flags);

    /* for kondara MNU/Linux 2.0 */
    if (scp->host->hostt->reset == NULL)
        scp->host->hostt->reset = cb31_reset;
//  printk("cb31_abort : end %d\n", ret);

    ASC_ASSERT(ret != ASC_ERROR);
    return ret;
} /* cb31_abort */

/*
 * cb31_reset()
 *
 * Reset the bus associated with the command 'scp'.
 *
 * This function runs its own thread. Interrupts must be blocked but
 * sleeping is allowed and no locking other than for host structures is
 * required. Returns SUCCESS or FAILED.
 */
int
cb31_reset(Scsi_Cmnd *scp, unsigned int flg)
{
    struct Scsi_Host     *shp;
    asc_board_t          *boardp;
    ASC_DVC_VAR          *asc_dvc_varp;
    unsigned long                flags;
    Scsi_Cmnd            *done_scp = NULL, *last_scp = NULL;
    Scsi_Cmnd            *tscp, *new_last_scp;
    int                  status;
    int                  ret = SUCCESS;
    int                  chipstat;

    ASC_DBG1(1, "cb31_reset: 0x%lx\n", (ulong) scp);

#ifdef ADVANSYS_STATS
    if (scp->host != NULL) {
        ASC_STATS(scp->host, reset);
    }
#endif /* ADVANSYS_STATS */

    if ((shp = scp->host) == NULL) {
        scp->result = HOST_BYTE(DID_ERROR);
        return FAILED;
    }

    boardp = ASC_BOARDP(shp);

    ASC_PRINT1("cb31_reset: board %d: SCSI bus reset started...\n",
        boardp->id);

    if (cb31_Insert == 0) {
        printk("cb31_reset: card is removed\n");
        scp->result = HOST_BYTE(DID_NO_CONNECT);
        scp->eh_state = FAILED;
        scp->host->in_recovery = 1;
        scp->device->online = FALSE;
        ret = SCSI_RESET_ERROR;
        restore_flags(flags);
        return(ret);
    } /* if (b31_Insert... */

    chipstat = AscGetChipStatus(boardp->dvc_var.asc_dvc_var.iop_base);
    if (chipstat & CSW_PCIERR) {
        printk("cb31_reset: detect PCIERR.\n");
        scp->result = HOST_BYTE(DID_ERROR);
        scp->host->in_recovery = 1;

        spin_lock_irqsave(&boardp->lock, flags);
        done_scp = asc_dequeue_list(&boardp->done, &last_scp, ASC_TID_ALL);
        if (done_scp != NULL) {
            done_scp->result = scp->result;
            done_scp->host->in_recovery = 1;
            done_scp->eh_state = scp->eh_state;
            done_scp->device = scp->device;
            done_scp->host->eh_wait = scp->host->eh_wait;
            asc_scsi_done_list(done_scp);
        }
        spin_unlock_irqrestore(&boardp->lock, flags);
        return FAILED;
    }
/* =================  2001-10-27 test */
    /*
     * Check for re-entrancy.
     */
    spin_lock_irqsave(&boardp->lock, flags);
    if (boardp->flags & ASC_HOST_IN_RESET) {
        spin_unlock_irqrestore(&boardp->lock, flags);
        return FAILED;
    }
    boardp->flags |= ASC_HOST_IN_RESET;
    spin_unlock_irqrestore(&boardp->lock, flags);

    /*
     * XXX - Remove this comment and the next line when SCSI mid-level
     * no longer acquires 'io_request_lock' before calling the SCSI
     * low-level reset entrypoint.
     */
    ASC_UNLOCK_IO_REQUEST_LOCK

    asc_dvc_varp = &boardp->dvc_var.asc_dvc_var;

    /*
     * Reset the chip and SCSI bus.
     */
    ASC_DBG(1, "cb31_reset: before AscInitAsc1000Driver()\n");
    status = AscInitAsc1000Driver(asc_dvc_varp);

    /* Refer to ASC_IERR_* defintions for meaning of 'err_code'. */
    if (asc_dvc_varp->err_code) {
        ASC_PRINT2(
            "cb31_reset: board %d: SCSI bus reset error: 0x%x\n",
            boardp->id, asc_dvc_varp->err_code);
        ret = FAILED;
    } else if (status) {
        ASC_PRINT2(
            "cb31_reset: board %d: SCSI bus reset warning: 0x%x\n",
            boardp->id, status);
    } else {
        ASC_PRINT1(
            "cb31_reset: board %d: SCSI bus reset successful.\n",
            boardp->id);
    }

    ASC_DBG(1, "cb31_reset: after AscInitAsc1000Driver()\n");

    /*
     * Acquire the board lock.
     */
    spin_lock_irqsave(&boardp->lock, flags);


    /* Board lock is held. */

    /*
     * Dequeue all board 'done' requests. A pointer to the last request
     * is returned in 'last_scp'.
     */
    done_scp = asc_dequeue_list(&boardp->done, &last_scp, ASC_TID_ALL);

    /*
     * Dequeue all board 'active' requests for all devices and set
     * the request status to DID_RESET. A pointer to the last request
     * is returned in 'last_scp'.
     */
    if (done_scp == NULL) {
        done_scp = asc_dequeue_list(&boardp->active, &last_scp, ASC_TID_ALL);
        for (tscp = done_scp; tscp; tscp = REQPNEXT(tscp)) {
            tscp->result = HOST_BYTE(DID_RESET);
        }
    } else {
        /* Append to 'done_scp' at the end with 'last_scp'. */
        ASC_ASSERT(last_scp != NULL);
        REQPNEXT(last_scp) = asc_dequeue_list(&boardp->active,
            &new_last_scp, ASC_TID_ALL);
        if (new_last_scp != NULL) {
            ASC_ASSERT(REQPNEXT(last_scp) != NULL);
            for (tscp = REQPNEXT(last_scp); tscp; tscp = REQPNEXT(tscp)) {
                tscp->result = HOST_BYTE(DID_RESET);
            }
            last_scp = new_last_scp;
        }
    }

    /*
     * Dequeue all 'waiting' requests and set the request status
     * to DID_RESET.
     */
    if (done_scp == NULL) {
        done_scp = asc_dequeue_list(&boardp->waiting, &last_scp, ASC_TID_ALL);
        for (tscp = done_scp; tscp; tscp = REQPNEXT(tscp)) {
            tscp->result = HOST_BYTE(DID_RESET);
        }
    } else {
        /* Append to 'done_scp' at the end with 'last_scp'. */
        ASC_ASSERT(last_scp != NULL);
        REQPNEXT(last_scp) = asc_dequeue_list(&boardp->waiting,
            &new_last_scp, ASC_TID_ALL);
        if (new_last_scp != NULL) {
            ASC_ASSERT(REQPNEXT(last_scp) != NULL);
            for (tscp = REQPNEXT(last_scp); tscp; tscp = REQPNEXT(tscp)) {
                tscp->result = HOST_BYTE(DID_RESET);
            }
            last_scp = new_last_scp;
        }
    }

    /* Save the time of the most recently completed reset. */
    boardp->last_reset = jiffies;

    /* Clear reset flag. */
    boardp->flags &= ~ASC_HOST_IN_RESET;

    /* Release the board. */
    spin_unlock_irqrestore(&boardp->lock, flags);

    /*
     * Complete all the 'done_scp' requests.
     */
    if (done_scp != NULL) {
        asc_scsi_done_list(done_scp);
    }

    /*
     * XXX - Remove this comment and the next line when SCSI mid-level
     * no longer acquires 'io_request_lock' before calling the SCSI
     * low-level reset entrypoint.
     */
    ASC_LOCK_IO_REQUEST_LOCK

    ASC_DBG1(1, "cb31_reset: ret %d\n", ret);

    return ret;
}  /* cb31_reset */

/*
 * cb31_biosparam()
 *
 * Translate disk drive geometry if the "BIOS greater than 1 GB"
 * support is enabled for a drive.
 *
 * ip (information pointer) is an int array with the following definition:
 * ip[0]: heads
 * ip[1]: sectors
 * ip[2]: cylinders
 */
int
cb31_biosparam(Disk *dp, kdev_t dep, int ip[])
{
    asc_board_t     *boardp;

    ASC_DBG(1, "cb31_biosparam: begin\n");
    ASC_STATS(dp->device->host, biosparam);
    boardp = ASC_BOARDP(dp->device->host);
    if ((boardp->dvc_var.asc_dvc_var.dvc_cntl &
         ASC_CNTL_BIOS_GT_1GB) && dp->capacity > 0x200000) {
            ip[0] = 255;
            ip[1] = 63;
    } else {
            ip[0] = 64;
            ip[1] = 32;
    }
    ip[2] = dp->capacity / (ip[0] * ip[1]);
    ASC_DBG(1, "cb31_biosparam: end\n");
    return 0;
} /* cb31_biosparam */

/*
 * cb31_setup()
 *
 * This function is called from init/main.c at boot time.
 * It it passed LILO parameters that can be set from the
 * LILO command line or in /etc/lilo.conf.
 *
 * It is used by the AdvanSys driver to either disable I/O
 * port scanning or to limit scanning to 1 - 4 I/O ports.
 * Regardless of the option setting EISA and PCI boards
 * will still be searched for and detected. This option
 * only affects searching for ISA and VL boards.
 *
 * If ADVANSYS_DEBUG is defined the driver debug level may
 * be set using the 5th (ASC_NUM_IOPORT_PROBE + 1) I/O Port.
 *
 * Examples:
 * 1. Eliminate I/O port scanning:
 *         boot: linux advansys=
 *       or
 *         boot: linux advansys=0x0
 * 2. Limit I/O port scanning to one I/O port:
 *        boot: linux advansys=0x110
 * 3. Limit I/O port scanning to four I/O ports:
 *        boot: linux advansys=0x110,0x210,0x230,0x330
 * 4. If ADVANSYS_DEBUG, limit I/O port scanning to four I/O ports and
 *    set the driver debug level to 2.
 *        boot: linux advansys=0x110,0x210,0x230,0x330,0xdeb2
 *
 * ints[0] - number of arguments
 * ints[1] - first argument
 * ints[2] - second argument
 * ...
 */
void
cb31_setup(char *str, int *ints)
{
    int    i;

    if (asc_iopflag == ASC_TRUE) {
        printk("RATOC CardBus SCSI: 'cb31_cb' LILO option may appear only once\n");
        return;
    }

    asc_iopflag = ASC_TRUE;

#ifdef CB31_IOPORT
    if (ints[0] > ASC_NUM_IOPORT_PROBE) {
#ifdef ADVANSYS_DEBUG
        if ((ints[0] == ASC_NUM_IOPORT_PROBE + 1) &&
            (ints[ASC_NUM_IOPORT_PROBE + 1] >> 4 == 0xdeb)) {
            asc_dbglvl = ints[ASC_NUM_IOPORT_PROBE + 1] & 0xf;
        } else {
#endif /* ADVANSYS_DEBUG */
            printk("RATOC CardBus SCSI: only %d I/O ports accepted\n",
                ASC_NUM_IOPORT_PROBE);
#ifdef ADVANSYS_DEBUG
        }
#endif /* ADVANSYS_DEBUG */
    }
#endif

#ifdef ADVANSYS_DEBUG
    ASC_DBG1(1, "cb31_setup: ints[0] %d\n", ints[0]);
    for (i = 1; i < ints[0]; i++) {
        ASC_DBG2(1, " ints[%d] %x", i, ints[i]);
    }
    ASC_DBG(1, "\n");
#endif /* ADVANSYS_DEBUG */

    for (i = 1; i <= ints[0] && i <= ASC_NUM_IOPORT_PROBE; i++) {
        asc_ioport[i-1] = ints[i];
        ASC_DBG2(1, "cb31_setup: asc_ioport[%d] %x\n",
            i - 1, asc_ioport[i-1]);
    }
}


/*
 * --- Loadable Driver Support
 */

// #ifdef MODULE
// Scsi_Host_Template driver_template = CB31;
// # include "scsi_module.c"
// #endif /* MODULE */


/*
 * --- Miscellaneous Driver Functions
 */

/*
 * First-level interrupt handler.
 *
 * For versions > v1.3.70, 'dev_id' is a pointer to the interrupting
 * adapter's asc_board_t. Because all boards are currently checked
 * for interrupts on each interrupt, 'dev_id' is not referenced. 'dev_id'
 * could be used to identify an interrupt passed to the AdvanSys driver,
 * which is for a device sharing an interrupt with an AdvanSys adapter.
 */
static void
cb31_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
    unsigned long   flags;
    int             i;
    asc_board_t     *boardp;
    Scsi_Cmnd       *done_scp = NULL, *last_scp = NULL;
    Scsi_Cmnd       *new_last_scp;

//  printk( "cb31_interrupt: begin\n");

    /*
     * Check for interrupts on all boards.
     * AscISR() will call asc_isr_callback().
     */
    for (i = 0; i < asc_board_count; i++) {
        boardp = ASC_BOARDP(asc_host[i]);
        ASC_DBG2(2, "cb31_interrupt: i %d, boardp %lx\n",
            i, (ulong) boardp)

         /* Card removed: 2001-10-18 */
        if (AscGetChipSignatureByte(asc_host[i]->io_port) == 0xff) {
            printk("cb31_interrupt: card is removed %d\n", i);
            /* 2001-10-22 */
            spin_lock_irqsave(&boardp->lock, flags);  /* 2001-10-27 */
            done_scp = asc_dequeue_list(&boardp->done, NULL,
                    ASC_TID_ALL);
            asc_scsi_done_list(done_scp);
            cb31_Insert = 0;
            cb31_Insert_boardp = boardp;
            /* 2001-10-22 */
            continue;
        }

        spin_lock_irqsave(&boardp->lock, flags);  /* 2001-10-25 */
        if (AscIsIntPending(asc_host[i]->io_port)) {
            ASC_STATS(asc_host[i], interrupt);
            ASC_DBG(1, "cb31_interrupt: before AscISR()\n");
            AscISR(&boardp->dvc_var.asc_dvc_var);
        }

        /*
         * Start waiting requests and create a list of completed requests.
         *
         * If a reset or abort request is being performed for the board,
         * the reset or abort handler will complete pending requests after
         * it has completed.
         */
        if ((boardp->flags & (ASC_HOST_IN_RESET | ASC_HOST_IN_ABORT)) == 0) {
            ASC_DBG2(1, "cb31_interrupt: done_scp %lx, last_scp %lx\n",
                (ulong) done_scp, (ulong) last_scp);

            /* Start any waiting commands for the board. */
            if (!ASC_QUEUE_EMPTY(&boardp->waiting)) {
                ASC_DBG(1, "cb31_interrupt: before asc_execute_queue()\n");
                asc_execute_queue(&boardp->waiting);
            }

             /*
              * Add to the list of requests that must be completed.
              *
              * 'done_scp' will always be NULL on the first iteration
              * of this loop. 'last_scp' is set at the same time as
              * 'done_scp'.
              */
            if (done_scp == NULL) {
                done_scp = asc_dequeue_list(&boardp->done, &last_scp,
                    ASC_TID_ALL);
            } else {
                ASC_ASSERT(last_scp != NULL);
                REQPNEXT(last_scp) = asc_dequeue_list(&boardp->done,
                    &new_last_scp, ASC_TID_ALL);
                if (new_last_scp != NULL) {
                    ASC_ASSERT(REQPNEXT(last_scp) != NULL);
                    last_scp = new_last_scp;
                }
            } // if (done_scp == NULL)
        } // if ((boardp->flags & (ASC_HOST_IN_RESET | ASC_HOST_IN_ABORT)) == 0)
        /* Interrupts could be enabled here. */
        spin_unlock_irqrestore(&boardp->lock, flags);
    } // for (i = 0; i < asc_board_count; i++)


    /*
     * It is possible for the request done function to re-enable
     * interrupts without confusing the driver. But here the
     * original flags aren't restored until all requests have been
     * completed.
     */
    asc_scsi_done_list(done_scp);

    ASC_DBG(1, "cb31_interrupt: end\n");
//  printk( "cb31_interrupt: end\n");
    return;
}  /* cb31_interrupt */

/*
 * Set the number of commands to queue per device for the
 * specified host adapter.
 */
static void
cb31_select_queue_depths(struct Scsi_Host *shp, Scsi_Device *devicelist)
{
    Scsi_Device        *device;
    asc_board_t        *boardp;

    boardp = ASC_BOARDP(shp);
    boardp->flags |= ASC_SELECT_QUEUE_DEPTHS;
    for (device = devicelist; device != NULL; device = device->next) {
        if (device->host != shp) {
            continue;
        }
        /*
         * Save a pointer to the device and set its initial/maximum
         * queue depth.
         */
        boardp->device[device->id] = device;
        device->queue_depth =
            boardp->dvc_var.asc_dvc_var.max_dvc_qng[device->id];
        ASC_DBG3(1, "cb31_select_queue_depths: shp %x, id %d, depth %d\n",
            (unsigned) shp, device->id, device->queue_depth);
    }
}   /* cb31_select_queue_depths */

/*
 * Function used only with polled I/O requests that are initiated by
 * advansys_command().
 */
static void
cb31_command_done(Scsi_Cmnd *scp)
{
    ASC_DBG1(1, "cb31_command_done: scp %x\n", (unsigned) scp);
    scp->SCp.Status = 1;
}

// #include "ascmcode.c"
#include "cb31_asc.c"

#if CONFIG_HOTPLUG
static int __devinit cb31_probe(struct pci_dev *pdev, const struct pci_device_id *id)
{
    unsigned char  chip_rev;
    unsigned short tmp16;
    char s[20];
    int base_addr;

//    printk(KERN_INFO "cb31_cb : cb31_probe\n");
      if (pci_enable_device(pdev))
              return -ENODEV;

      /* disabe all powermanagement */
      pci_write_config_dword(pdev, PCI_POWERMGMT, 0x0000);

      pci_set_master(pdev);

      pci_read_config_word(pdev, PCI_STATUS, &tmp16);
      pci_write_config_word(pdev, PCI_STATUS, &tmp16);

      pci_read_config_byte(pdev, PCI_REVISION_ID, &chip_rev);

      base_addr = pci_resource_start(pdev, 0);

      if (!request_region(base_addr, 256, "cb31_cb")) {
        printk(KERN_ERR "cb31_probe: failed to allocate io-region\n");
        return -ENODEV;
      }

      sprintf(s, "cb31_setup");
      cb31_setup(s, ints);

      scsi_register_module(MODULE_SCSI_HA, &driver_template);
      if (! driver_template.present) {
        scsi_unregister_module(MODULE_SCSI_HA, &driver_template);
        release_region(base_addr, 256);
        return -ENODEV;
      }

      MOD_INC_USE_COUNT;
      return 0;
}  /* cb31_probe */

static void __devexit cb31_remove(struct pci_dev *pdev)
{
    unsigned long        flags;
    Scsi_Cmnd            *scp = NULL;

//  printk("cb31_cb: cb31_remove\n");
    if (cb31_Insert_boardp != NULL) {
            spin_lock_irqsave(&cb31_Insert_boardp->lock, flags);  /* 2001-10-27 */
            scp = asc_dequeue_list(&cb31_Insert_boardp->waiting, NULL,
                    ASC_TID_ALL);
            if (scp != NULL) {
                //    printk("cb31_remove: asc_scsi_done_list\n");
                DvcSleepMilliSecond(500L);
                scp->result = HOST_BYTE(DID_NO_CONNECT);
                scp->eh_state = FAILED;
                scp->device->online = FALSE;
                asc_scsi_done_list(scp);
                spin_unlock_irqrestore(&cb31_Insert_boardp->lock, flags);
            } else {
                //    printk("cb31_remove: done_scp is NULL\n");
                spin_unlock_irqrestore(&cb31_Insert_boardp->lock, flags);
            }
    }
    scsi_unregister_module(MODULE_SCSI_HA, &driver_template);
    MOD_DEC_USE_COUNT;
    printk("cb31_cb: cb31_remove : finish\n");
    return;
}  /* cb31_remove */

#else  // CONFIG_HOTPLUG

/*===================================================================*/

static dev_node_t *cb31_attach(dev_locator_t *loc)
{
    u_int  io;
    u_char bus, devfn, irq;
    dev_node_t *node;
    char s[20];

    if (loc->bus != LOC_PCI) {
	printk(KERN_INFO "cb31_cb : not CardBus (PCI) \n");
	return NULL;
    }

    bus = loc->b.pci.bus; devfn = loc->b.pci.devfn;
    printk(KERN_INFO "cb31_cb : attach(bus %d, fn %d)\n", bus, devfn);
    pcibios_read_config_dword(bus, devfn, PCI_BASE_ADDRESS_0, &io);
    pcibios_read_config_byte(bus, devfn, PCI_INTERRUPT_LINE, &irq);
    printk(KERN_INFO "cb31_cb : io = 0x%04X irq = %d\n", io, irq);

    sprintf(s, "cb31_setup");
    cb31_setup(s, ints);
    printk(KERN_INFO "cb31_cb : cb31_setup\n");
    scsi_register_module(MODULE_SCSI_HA, &driver_template);
    if (! driver_template.present) {
    printk(KERN_INFO "cb31_cb : scsi_register_module error !!!\n");
	scsi_unregister_module(MODULE_SCSI_HA, &driver_template);
	return NULL;
    }

    printk(KERN_INFO "cb31_cb : scsi_register_module\n");

    node = cb31_config(loc);
    if (node != NULL) {
	MOD_INC_USE_COUNT;
        cb31_Insert = 1;  /* 2001-10-22 */
    }

    return node;
}  /* cb31_attach */

static void cb31_detach(dev_node_t *node)
{
    DEBUG(0, "cb31_detach \n");
    scsi_unregister_module(MODULE_SCSI_HA, &driver_template);
    printk(KERN_INFO "cb31_cb : scsi_unregister_module\n");
    MOD_DEC_USE_COUNT;
    if (node) kfree(node);
    cb31_Insert = 0; /* 2001-10-22 */
    return;
}  /* cb31_detach */

/*===================================================================*/


static dev_node_t *cb31_config(dev_locator_t *loc)
{
    Scsi_Device *dev;
    dev_node_t  *node;
    int n = 0;
    struct Scsi_Host *host;

// printk("cb31_config:\n");

   node = kmalloc(7 * sizeof(dev_node_t), GFP_KERNEL);
    for (host = scsi_hostlist; host; host = host->next)
	if (host->hostt == &driver_template)
	    for (dev = host->host_queue; dev; dev = dev->next) {
		ulong arg[2], id;
		kernel_scsi_ioctl(dev, SCSI_IOCTL_GET_IDLUN, arg);
		id = (arg[0]&0x0f) + ((arg[0]>>4)&0xf0) +
		    ((arg[0]>>8)&0xf00) + ((arg[0]>>4)&0xf000);
    printk(KERN_INFO "cb31_cb : kernel_scsi_ioctl id = %ld\n", id);
		node->minor = 0;
		switch (dev->type) {
		case TYPE_TAPE:
		    node->major = SCSI_TAPE_MAJOR;
		    sprintf(node->dev_name, "st#%04lx", id);
		    break;
		case TYPE_DISK:
		case TYPE_MOD:
		    node->major = SCSI_DISK0_MAJOR;
		    sprintf(node->dev_name, "sd#$04lx", id);
		    break;
		case TYPE_ROM:
		case TYPE_WORM:
		    node->major = SCSI_CDROM_MAJOR;
		    sprintf(node->dev_name, "sr#$04lx", id);
		    break;
		default:
		    node->major = SCSI_GENERIC_MAJOR;
		    sprintf(node->dev_name, "sg#$04lx", id);
		    break;
		}
                if (n) node[n-1].next = &node[n];
		n++;
	    }
    if (n == 0) {
	printk(KERN_INFO "cb31_cb: no SCSI devices found\n");
	kfree(node);
	node = NULL;
    } else {
	printk(KERN_INFO "cb31_cb: %d SCSI devices found\n", n);
	node[n-1].next = NULL;
    }

    return node;
} /* cb31_config */

#endif  // CONFIG_HOTPLUG
/*===================================================================*/


#if LINUX_VERSION_CODE < CVT_LINUX_VERSION(2,4,0)
int init_module(void)
#else
static int __init init_cb31_cb(void)
#endif
{
    printk(KERN_INFO "cb31_cb : %s start\n", version);

#if CONFIG_HOTPLUG
    pci_register_driver(&cb31_pci_ops);
#else
    register_driver(&cb31_ops);
#endif
    return 0;
}

#if LINUX_VERSION_CODE < CVT_LINUX_VERSION(2,4,0)
void cleanup_module(void)
#else
static void __exit exit_cb31_cb(void)
#endif
{
    if (cb31_Insert) {
        cb31_release(asc_host[0]);
    }
#if CONFIG_HOTPLUG
    pci_unregister_driver(&cb31_pci_ops);
#else
    unregister_driver(&cb31_ops);
#endif
    printk(KERN_INFO "cb31_cb : quit\n");
}

#if LINUX_VERSION_CODE >= CVT_LINUX_VERSION(2,4,0)
module_init(init_cb31_cb);
module_exit(exit_cb31_cb);
#endif


// cb31main.c

