/* $Id: cb31_cb.h,v 1.18 2001/05/25 21:23:49 bobf Exp bobf $ */

/*
 * advansys.h - Linux Host Driver for AdvanSys SCSI Adapters
 *
 * Copyright (c) 1995-1998 Advanced System Products, Inc.
 * All Rights Reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that redistributions of source
 * code retain the above copyright notice and this comment without
 * modification.
 *
 * There is an AdvanSys Linux WWW page at:
 *  http://www.advansys.com/linux.html
 *
 * The latest version of the AdvanSys driver is available at:
 *  ftp://ftp.advansys.com/pub/linux
 *
 * Please send questions, comments, bug reports to:
 *  bobf@advansys.com (Bob Frey)
 */

#ifndef _ADVANSYS_H
#define _ADVANSYS_H

/* Convert Linux Version, Patch-level, Sub-level to LINUX_VERSION_CODE. */
#define CVT_LINUX_VERSION(V, P, S)    (((V) * 65536) + ((P) * 256) + (S))
// #define ASC_LINUX_VERSION(V, P, S)    (((V) * 65536) + ((P) * 256) + (S))

#ifndef LINUX_VERSION_CODE
#include <linux/version.h>
#endif /* LINUX_VERSION_CODE */

#define ASC_LINUX_KERNEL22 (LINUX_VERSION_CODE < ASC_LINUX_VERSION(2,4,0))
#define ASC_LINUX_KERNEL24 (LINUX_VERSION_CODE >= ASC_LINUX_VERSION(2,4,0))

#include "cb31_asc.h"

#if LINUX_VERSION_CODE < CVT_LINUX_VERSION(2,4,0)
#include <../../drivers/scsi/sd.h>
#else  // < 2.4.0
#include "../sd.h"
#endif // >=2.4.0

/*
 * Scsi_Host_Template function prototypes.
 */
int cb31_detect(Scsi_Host_Template *);
int cb31_release(struct Scsi_Host *);
const char *cb31_info(struct Scsi_Host *);
int cb31_command(Scsi_Cmnd *);
int cb31_queuecommand(Scsi_Cmnd *, void (* done)(Scsi_Cmnd *));
int cb31_abort(Scsi_Cmnd *);
int cb31_reset(Scsi_Cmnd *, unsigned int);

int cb31_biosparam(Disk *, kdev_t, int[]);

#if LINUX_VERSION_CODE < CVT_LINUX_VERSION(2,3,27) // 2001-05-31
extern struct proc_dir_entry proc_scsi_cb31;
#endif
int cb31_proc_info(char *, char **, off_t, int, int, int);

/* init/main.c setup function */
void cb31_setup(char *, int *);

#define cb31_REVID "RATOC CB31 SCSI device driver; Revision: 0.10"
/*
 * AdvanSys Host Driver Scsi_Host_Template (struct SHT) from hosts.h.
 */

#if LINUX_VERSION_CODE < CVT_LINUX_VERSION(2,3,27)
#define CB31 { \
    proc_dir:     &proc_scsi_cb31, \
    proc_info:    cb31_proc_info, \
    name:         cb31_REVID, \
    detect:       cb31_detect, \
    release:      cb31_release, \
    info:         cb31_info, \
    command:      cb31_command, \
    queuecommand: cb31_queuecommand, \
    abort:        cb31_abort, \
    reset:        cb31_reset, \
    bios_param:    cb31_biosparam, \
    /* \
     * Because the driver may control an ISA adapter 'unchecked_isa_dma' \
     * must be set. The flag will be cleared in advansys_detect for non-ISA \
     * adapters. Refer to the comment in scsi_module.c for more information. \
     */ \
    unchecked_isa_dma: 1, \
    /* \
     * All adapters controlled by this driver are capable of large \
     * scatter-gather lists. According to the mid-level SCSI documentation \
     * this obviates any performance gain provided by setting \
     * 'use_clustering'. But empirically while CPU utilization is increased \
     * by enabling clustering, I/O throughput increases as well. \
     */ \
    use_clustering: ENABLE_CLUSTERING, \
}
#else /* version >= v2.3.27 */
#define CB31 { \
    next:         NULL,	\
    module:       NULL,	\
    proc_name:    "CB31", \
    proc_info:    NULL, \
    name:         cb31_REVID, \
    detect:       cb31_detect, \
    release:      cb31_release, \
    info:         cb31_info, \
    command:      cb31_command, \
    queuecommand: cb31_queuecommand, \
    eh_strategy_handler: NULL, \
    eh_abort_handler: NULL, \
    eh_device_reset_handler: NULL, \
    eh_bus_reset_handler: cb31_reset, \
    eh_host_reset_handler: NULL, \
    abort:        cb31_abort,  \
/*    reset:        cb31_reset, */ \
    slave_attach: NULL, \
    bios_param:   cb31_biosparam, \
    can_queue:	  1,                /* int can_queue */ \
    this_id:	  1,                /* int this_id */ \
    sg_tablesize: SG_ALL,           /* short unsigned int sg_tablesize */ \
    cmd_per_lun:  1,                /* short cmd_per_lun */ \
    present:	  0,                /* unsigned char present */    \
    /* \
     * Because the driver may control an ISA adapter 'unchecked_isa_dma' \
     * must be set. The flag will be cleared in advansys_detect for non-ISA \
     * adapters. Refer to the comment in scsi_module.c for more information. \
     */ \
    unchecked_isa_dma: 1, \
    /* \
     * All adapters controlled by this driver are capable of large \
     * scatter-gather lists. According to the mid-level SCSI documentation \
     * this obviates any performance gain provided by setting \
     * 'use_clustering'. But empirically while CPU utilization is increased \
     * by enabling clustering, I/O throughput increases as well. \
     */ \
    use_clustering: ENABLE_CLUSTERING, \
  use_new_eh_code: 0 \
}
#endif   /* version v2.3.27 */

#endif /* _ADVANSYS_H */
