/**************************************************************************
 * CB32 (Inic 850) device driver for Linux.
 *
 * Copyright (c) 1994-1998 Initio Corporation
 * Copyright (c) 2000-2001 RATOC Systems, Inc.
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * --------------------------------------------------------------------------
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions, and the following disclaimer,
 *    without modification, immediately at the beginning of the file.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * Where this Software is combined with software released under the terms of 
 * the GNU Public License ("GPL") and the terms of the GPL would require the 
 * combined work to also be released under the terms of the GPL, the terms
 * and conditions of this License will apply in addition to those of the
 * GPL with the exception of any terms or conditions of this License that
 * conflict with, or are expressly prohibited by, the GPL.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 **************************************************************************
 * 
 * module: cb32i850.c
 * DESCRIPTION:
 * 	This is the Linux low-level SCSI driver for RATOC CB32 SCSI host
 * 	adapters
 * 09/24/1998 hl - v1.02 initial production release.
 * 12/19/1998 bv - v1.02a Use spinlocks for 2.1.95 and up.
 * 05/10/2000 ko - Inic 850 new 
 * 07/07/2000 sk - CB32 new
 * 02/23/2001 sk - v0.50 for kernel 2.4.x.
 * 05/21/2001 sk - v0.51 Fixed scsi abort.
 * 10/09/2001 sk - v0.60 for HotPlug. add cb32_release
 **************************************************************************/

#define CVT_LINUX_VERSION(V,P,S)        (V * 65536 + P * 256 + S)

#ifndef LINUX_VERSION_CODE
#include <linux/version.h>
#endif

#ifdef MODULE
#include <linux/module.h>
#endif // MODULE

#if LINUX_VERSION_CODE < CVT_LINUX_VERSION(2,4,0)
#include <pcmcia/config.h>
#include <pcmcia/k_compat.h>
#endif // 2.4.0

#if LINUX_VERSION_CODE >= CVT_LINUX_VERSION(1,3,0)
#include <stdarg.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/ioport.h>
#include <linux/delay.h>
#include <linux/sched.h>
#if LINUX_VERSION_CODE <= CVT_LINUX_VERSION(2,1,92)
#include <linux/bios32.h>
#endif   // 2.1.92
#include <linux/pci.h>
#if LINUX_VERSION_CODE < CVT_LINUX_VERSION(2,3,27)
#include <linux/proc_fs.h>
#endif   // 2.3.27
#if LINUX_VERSION_CODE >= CVT_LINUX_VERSION(2,1,23)
#include <linux/init.h>
#endif   // 2.1.23
#include <linux/blk.h>
#if LINUX_VERSION_CODE >= CVT_LINUX_VERSION(2,3,17)
#include <linux/spinlock.h>
#else    // 2.3.17
#if LINUX_VERSION_CODE >= CVT_LINUX_VERSION(2,1,95)
#include <asm/spinlock.h>
#endif     // 2.1.95
#endif   // 2.3.17
// #include "sd.h"
#if LINUX_VERSION_CODE < CVT_LINUX_VERSION(2,4,0)
#include "../../drivers/scsi/scsi.h"
#include "../../drivers/scsi/hosts.h"
#else  // 2.4.0
#include "../scsi.h"
#include "../hosts.h"
#endif // 2.4.0
#include "cb32i850.h"
#include <linux/stat.h>
#include <linux/malloc.h>
#include <linux/config.h>

#else // 1.3.0

#include <linux/kernel.h>
#include <linux/head.h>
#include <linux/types.h>
#include <linux/string.h>
#include <linux/ioport.h>

#include <linux/sched.h>
#include <linux/proc_fs.h>
#include <asm/system.h>
#include <asm/io.h>
#include "../block/blk.h"
#include "drivers/scsi/scsi.h"
#include "sd.h"
#include "drivers/scsi/hosts.h"
#include <linux/malloc.h>
#include "cb32i850.h"
#endif // 1.3.0

#include <pcmcia/driver_ops.h>

#ifdef PCMCIA_DEBUG
static int pc_debug = PCMCIA_DEBUG;
MODULE_PARM(pc_debug, "i");
#define DEBUG(n, args...) if (pc_debug>(n)) printk(KERN_DEBUG args)
#else
#define DEBUG(n, args...)
#endif

// 02/20/2001
#if LINUX_VERSION_CODE >= CVT_LINUX_VERSION(2,4,0)
static Scsi_Host_Template driver_template = CB32;
// #include "../scsi_module.c"
#else
#ifdef MODULE
Scsi_Host_Template driver_template = CB32;
// #include "scsi_module.c"
#endif
#endif

#define ORC_RDWORD(x,y)         (short)(inl((int)((ULONG)((ULONG)x+(UCHAR)y)) ))

char *cb32_Copyright = "Copyright (C) 1998-2001";
char *cb32_InitioName = "by Initio Corporation/RATOC";
char *cb32_ProductName = "CB32";
char *cb32_Version = "v0.60";

// 02/20/2001
#if LINUX_VERSION_CODE >= CVT_LINUX_VERSION(1,3,0) && LINUX_VERSION_CODE < CVT_LINUX_VERSION(2,3,27)

#define PROC_SCSI_CB32 PROC_SCSI_INIA100

struct proc_dir_entry proc_scsi_cb32 =
{
	PROC_SCSI_CB32, 4, "CB32",
	S_IFDIR | S_IRUGO | S_IXUGO, 2,
	0, 0, 0, NULL, NULL, NULL, NULL, NULL, NULL, NULL
};
#endif

static int ints[3] = { 0, 0, 0};

static dev_node_t *cb32_attach(dev_locator_t *loc);
static void cb32_detach(dev_node_t *node);

struct driver_operations cb32_ops = {
    "cb32_cb", cb32_attach, NULL, NULL, cb32_detach
};


/* set by cb32_setup according to the command line */
static int setup_called = 0;
static int orc_num_ch = MAX_SUPPORTED_ADAPTERS;		/* Maximum 4 adapters           */

/* ---- INTERNAL VARIABLES ---- */
#define NUMBER(arr)     (sizeof(arr) / sizeof(arr[0]))
static char *setup_str = (char *) NULL;

#if LINUX_VERSION_CODE >= CVT_LINUX_VERSION(1,3,0)
static void cb32_intr0(int irq, void *dev_id, struct pt_regs *);
static void cb32_intr1(int irq, void *dev_id, struct pt_regs *);
static void cb32_intr2(int irq, void *dev_id, struct pt_regs *);
static void cb32_intr3(int irq, void *dev_id, struct pt_regs *);
static void cb32_intr4(int irq, void *dev_id, struct pt_regs *);
static void cb32_intr5(int irq, void *dev_id, struct pt_regs *);
static void cb32_intr6(int irq, void *dev_id, struct pt_regs *);
static void cb32_intr7(int irq, void *dev_id, struct pt_regs *);
#else
static void cb32_intr0(int irq, struct pt_regs *);
static void cb32_intr1(int irq, struct pt_regs *);
static void cb32_intr2(int irq, struct pt_regs *);
static void cb32_intr3(int irq, struct pt_regs *);
static void cb32_intr4(int irq, struct pt_regs *);
static void cb32_intr5(int irq, struct pt_regs *);
static void cb32_intr6(int irq, struct pt_regs *);
static void cb32_intr7(int irq, struct pt_regs *);
#endif

static void cb32_panic(char *msg);
void cb32SCBPost(BYTE * pHcb, BYTE * pScb);

/* ---- EXTERNAL VARIABLES ---- */
extern int Addcb32_into_Adapter_table(WORD, WORD, BYTE, BYTE, BYTE);
extern void init_cb32Adapter_table(void);
extern ORC_SCB *orc_alloc_scb(ORC_HCS * hcsp);
extern void orc_exec_scb(ORC_HCS * hcsp, ORC_SCB * scbp);
extern void orc_release_scb(ORC_HCS * hcsp, ORC_SCB * scbp);
extern void orc_interrupt(ORC_HCS * hcsp);
extern int orc_device_reset(ORC_HCS * pHCB, ULONG SCpnt, unsigned int target, unsigned int ResetFlags);
extern int orc_reset_scsi_bus(ORC_HCS * pHCB);
extern int abort_SCB(ORC_HCS * hcsp, ORC_SCB * pScb);
extern int orc_abort_srb(ORC_HCS * hcsp, ULONG SCpnt);
extern void get_orcPCIConfig(ORC_HCS * pCurHcb, int ch_idx);
extern int init_orchid(ORC_HCS * hcsp);

extern int orc_num_scb;
extern ORC_HCS orc_hcs[];

/*============================== for HotPlug =============================*/
#if CONFIG_HOTPLUG
static struct pci_device_id cb32_pci_table[] __devinitdata = {
	{ 0x1195, 0x0851, 0x1195, 0x0851, },
	{ 0, },
};
MODULE_DEVICE_TABLE(pci, cb32_pci_table);

static int cb32_probe(struct pci_dev *pdev, const struct pci_device_id *id);
static void cb32_remove(struct pci_dev *pdev);

static struct pci_driver cb32_pci_ops = {
	name:		"cb32_cb",
	id_table:	cb32_pci_table,
	probe:		cb32_probe,
	remove:		cb32_remove,
	suspend:	NULL,
	resume:		NULL
};

struct Scsi_Host *cb32_dev_id;

#define PCI_POWERMGMT    0x40  // PCI register for PM

#endif

/*****************************************************************************
 Function name  : cb32AppendSRBToQueue
 Description    : This function will push current request into save list
 Input          : pSRB  -       Pointer to SCSI request block.
		  pHCB  -       Pointer to host adapter structure
 Output         : None.
 Return         : None.
*****************************************************************************/
static void cb32AppendSRBToQueue(ORC_HCS * pHCB, Scsi_Cmnd * pSRB)
{
	ULONG flags;

#if LINUX_VERSION_CODE >= CVT_LINUX_VERSION(2,1,95)
	spin_lock_irqsave(&(pHCB->pSRB_lock), flags);
#else
	save_flags(flags);
	cli();
#endif

	pSRB->next = NULL;	/* Pointer to next */
	if (pHCB->pSRB_head == NULL)
		pHCB->pSRB_head = pSRB;
	else
		pHCB->pSRB_tail->next = pSRB;	/* Pointer to next */
	pHCB->pSRB_tail = pSRB;
#if LINUX_VERSION_CODE >= CVT_LINUX_VERSION(2,1,95)
	spin_unlock_irqrestore(&(pHCB->pSRB_lock), flags);
#else
	restore_flags(flags);
#endif
	return;
}

/*****************************************************************************
 Function name  : cb32PopSRBFromQueue
 Description    : This function will pop current request from save list
 Input          : pHCB  -       Pointer to host adapter structure
 Output         : None.
 Return         : pSRB  -       Pointer to SCSI request block.
*****************************************************************************/
static Scsi_Cmnd *cb32PopSRBFromQueue(ORC_HCS * pHCB)
{
	Scsi_Cmnd *pSRB;
	ULONG flags;
#if LINUX_VERSION_CODE >= CVT_LINUX_VERSION(2,1,95)
	spin_lock_irqsave(&(pHCB->pSRB_lock), flags);
#else
	save_flags(flags);
	cli();
#endif

	if ((pSRB = (Scsi_Cmnd *) pHCB->pSRB_head) != NULL) {
		pHCB->pSRB_head = pHCB->pSRB_head->next;
		pSRB->next = NULL;
	}
#if LINUX_VERSION_CODE >= CVT_LINUX_VERSION(2,1,95)
	spin_unlock_irqrestore(&(pHCB->pSRB_lock), flags);
#else
	restore_flags(flags);
#endif
	return (pSRB);
}

/*****************************************************************************
 Function name  : cb32_setup
 Description    : 
 Input          : pHCB  -       Pointer to host adapter structure
 Output         : None.
 Return         : pSRB  -       Pointer to SCSI request block.
*****************************************************************************/
void cb32_setup(char *str, int *ints)
{
	if (setup_called)
		cb32_panic("cb32_cb: cb32_setup called twice.\n");

	setup_called = ints[0];
	setup_str = str;
}

/*****************************************************************************
 Function name	: orc_ReturnNumberOfAdapters
 Description	: This function will scan PCI bus to get all Orchid card
 Input		: None.
 Output		: None.
 Return		: SUCCESSFUL	- Successful scan
		  ohterwise	- No drives founded
*****************************************************************************/
int orc_ReturnNumberOfAdapters(void)
{
    unsigned int i, iAdapters;

    iAdapters = 0;
	/*
	 * PCI-bus probe.
	 */
    if (pcibios_present()) {
	struct {
		unsigned short vendor_id;
		unsigned short device_id;
	} const cb32_pci_devices[] =
	{
		{ORC_VENDOR_ID, I920_DEVICE_ID},
		{ORC_VENDOR_ID, ORC_DEVICE_ID}
	};

	unsigned int dRegValue;
	unsigned short command;
	WORD wBIOS, wBASE;
	BYTE bPCIBusNum, bInterrupt, bPCIDeviceNum;

#ifdef MMAPIO
	unsigned long page_offset, base;
#endif

#if LINUX_VERSION_CODE > CVT_LINUX_VERSION(2,1,92)
	struct pci_dev *pdev = NULL;
#else
	int index;
	unsigned char pci_bus, pci_devfn;
#endif

	bPCIBusNum = 0;
	bPCIDeviceNum = 0;
	init_cb32Adapter_table();
	for (i = 0; i < NUMBER(cb32_pci_devices); i++) {
#if LINUX_VERSION_CODE > CVT_LINUX_VERSION(2,1,92)
	    pdev = NULL;
	    while ((pdev = pci_find_device(cb32_pci_devices[i].vendor_id,
					cb32_pci_devices[i].device_id,
						       pdev)))
#else
		index = 0;
		while (!(pcibios_find_device(cb32_pci_devices[i].vendor_id,
					cb32_pci_devices[i].device_id,
					 index++, &pci_bus, &pci_devfn)))
#endif
		{
		    if (iAdapters >= MAX_SUPPORTED_ADAPTERS)
			break;	/* Never greater than maximum   */

			if (i == 0) {
				/*
				   printk("cb32_cb: The RAID controller is not supported by\n");
				   printk("cb32_cb:         this driver, we are ignoring it.\n");
				 */
			} else {
				/*
				 * Read sundry information from PCI BIOS.
				 */
#if LINUX_VERSION_CODE > CVT_LINUX_VERSION(2,1,92)
			    bPCIBusNum = pdev->bus->number;
			    bPCIDeviceNum = pdev->devfn;

#if LINUX_VERSION_CODE < CVT_LINUX_VERSION(2,3,13)	// 02/20/2001
			    dRegValue = pdev->base_address[0]; 
#else
			    dRegValue = pci_resource_start(pdev,0);
#endif
			    if (dRegValue == -1) {	/* Check return code  */
				printk("\n\rcb32_cb: orchid read configuration error.\n");
				return (0);	/* Read configuration space error  */
			    }
				/* <02> read from base address + 0x50 offset to get the wBIOS balue. */
			    wBASE = (WORD) dRegValue;

				/* Now read the interrupt line  */
			    dRegValue = pdev->irq;
			    bInterrupt = dRegValue & 0xFF;	/* Assign interrupt line      */
			    pci_read_config_word(pdev, PCI_COMMAND, &command);
			    pci_write_config_word(pdev, PCI_COMMAND,
			      command | PCI_COMMAND_MASTER | PCI_COMMAND_IO);

#else
			    bPCIBusNum = pci_bus;
			    bPCIDeviceNum = pci_devfn;
			    pcibios_read_config_dword(pci_bus, pci_devfn,
					 PCI_BASE_ADDRESS_0, &dRegValue);
			    if (dRegValue == -1) {	/* Check return code */
				printk("\n\rcb32_cb: Orchid read configuration error.\n");
				return (0);	/* Read configuration space error  */
			    }
				/* <02> read from base address + 0x50 offset to get the wBIOS balue. */
			    wBASE = (WORD) dRegValue;

				/* Now read the interrupt line  */
			    pcibios_read_config_dword(pci_bus, pci_devfn,
					 PCI_INTERRUPT_LINE, &dRegValue);
			    bInterrupt = dRegValue & 0xFF;	/* Assign interrupt line      */
			    pcibios_read_config_word(pci_bus, pci_devfn, PCI_COMMAND, &command);
			    pcibios_write_config_word(pci_bus, pci_devfn, PCI_COMMAND,
				 command | PCI_COMMAND_MASTER | PCI_COMMAND_IO);
#endif
			    wBASE &= PCI_BASE_ADDRESS_IO_MASK;
			    wBIOS = ORC_RDWORD(wBASE, 0x50);

#ifdef MMAPIO
			    base = wBASE & PAGE_MASK;
			    page_offset = wBASE - base;

				/*
				 * replace the next line with this one if you are using 2.1.x:
				 * temp_p->maddr = ioremap(base, page_offset + 256);
				 */
#if LINUX_VERSION_CODE >= CVT_LINUX_VERSION(2,1,0)
			    wBASE = ioremap(base, page_offset + 256);
#else
			    wBASE = (WORD) vremap(base, page_offset + 256);
#endif
			    if (wBASE) {
				wBASE += page_offset;
			    }
#endif

			    if (Addcb32_into_Adapter_table(wBIOS, wBASE, bInterrupt, bPCIBusNum,
					    bPCIDeviceNum) == SUCCESSFUL)
				iAdapters++;
			    }
			}	/* while(pdev=....) */
		}		/* for PCI_DEVICES */
	}			/* PCI BIOS present */
	return (iAdapters);
}

/*****************************************************************************
 Function name  : cb32_detect
 Description    : 
 Input          : pHCB  -       Pointer to host adapter structure
 Output         : None.
 Return         : pSRB  -       Pointer to SCSI request block.
*****************************************************************************/
int cb32_detect(Scsi_Host_Template * tpnt)
{
    ORC_HCS *pHCB;
    struct Scsi_Host *hreg, *hreg2;	/* 02/21/2001		*/
    U32 sz;
    U32 i;			/* 01/14/98                     */
    int ok = 0, iAdapters;
    ULONG dBiosAdr;
    BYTE *pbBiosAdr;

#if LINUX_VERSION_CODE >= CVT_LINUX_VERSION(1,3,0)
#if LINUX_VERSION_CODE < CVT_LINUX_VERSION(2,3,27)
	tpnt->proc_dir = &proc_scsi_cb32;
#else
	tpnt->proc_name = "CB32";
#endif
#endif
    if (setup_called) {
	/* Setup by cb32_setup          */
	printk("cb32_cb: processing commandline: ");
    }
	/* Get total number of adapters in the motherboard */
    iAdapters = orc_ReturnNumberOfAdapters();

    printk("cb32_cb: Total Adapters = %d\n", iAdapters);
    if (iAdapters == 0)	/* If no orc founded, return */
	return (0);

    orc_num_ch = (iAdapters > orc_num_ch) ? orc_num_ch : iAdapters;
    orc_num_scb = ORC_MAXQUEUE;

    /* clear the memory needed for HCS */
    i = orc_num_ch * sizeof(ORC_HCS);
	/* Initialize orc_hcs 0 */
    memset((unsigned char *) &orc_hcs[0], 0, i);

#if 0
    printk("orc_num_scb= %x orc_num_ch= %x hcsize= %x scbsize= %x escbsize= %x\n",
       orc_num_scb, orc_num_ch, sizeof(ORC_HCS), sizeof(ORC_SCB), sizeof(ESCB));
#endif

    for (i = 0, pHCB = &orc_hcs[0];	/* Get pointer for control block */
	     i < orc_num_ch;
	     i++, pHCB++) {

	pHCB->pSRB_head = NULL;		/* Initial SRB save queue       */
	pHCB->pSRB_tail = NULL;		/* Initial SRB save queue       */
#if LINUX_VERSION_CODE >= CVT_LINUX_VERSION(2,1,95)
	pHCB->pSRB_lock = SPIN_LOCK_UNLOCKED; /* SRB save queue lock */
#endif
	/* Get total memory needed for SCB */
	sz = orc_num_scb * sizeof(ORC_SCB);
	if ((pHCB->HCS_virScbArray = (PVOID) kmalloc(sz, GFP_ATOMIC | GFP_DMA)) == NULL) {
		printk("cb32_cb: SCB memory allocation error\n");
		return (0);
	}
	memset((unsigned char *) pHCB->HCS_virScbArray, 0, sz);
	pHCB->HCS_physScbArray = (U32) VIRT_TO_BUS(pHCB->HCS_virScbArray);

	/* Get total memory needed for ESCB */
	sz = orc_num_scb * sizeof(ESCB);
	if ((pHCB->HCS_virEscbArray = (PVOID) kmalloc(sz, GFP_ATOMIC | GFP_DMA)) == NULL) {
		printk("cb32_cb: ESCB memory allocation error\n");
		return (0);
	}
	memset((unsigned char *) pHCB->HCS_virEscbArray, 0, sz);
	pHCB->HCS_physEscbArray = (U32) VIRT_TO_BUS(pHCB->HCS_virEscbArray);

	request_region(pHCB->HCS_Base, 0x100, "cb32");	/* Register */
	get_orcPCIConfig(pHCB, i);

#ifdef HAVE_BIOS
	dBiosAdr = pHCB->HCS_BIOS;
	dBiosAdr = (dBiosAdr << 4);

#if LINUX_VERSION_CODE >= CVT_LINUX_VERSION(1,3,0)
	pbBiosAdr = phys_to_virt(dBiosAdr);
#endif
#endif

	if (init_orchid(pHCB)) {	/* Initial orchid chip    */
		printk("cb32_cb: initial orchid fail!!\n");
		return (0);
	}
	hreg = scsi_register(tpnt, sizeof(ORC_HCS));
	if (hreg == NULL) {
		printk("Invalid scsi_register pointer.\n");
	}
	hreg->io_port = pHCB->HCS_Base;
	hreg->n_io_port = 0xff;
	hreg->can_queue = orc_num_scb;	/* 03/05/98    */

#if LINUX_VERSION_CODE >= CVT_LINUX_VERSION(1,3,0)
	hreg->unique_id = pHCB->HCS_Base;
	hreg->max_id = pHCB->HCS_MaxTar;
#endif

	hreg->max_lun = 32;	/* 10/21/97                     */
/*
   hreg->max_lun = 8;
   hreg->max_channel = 1;
 */
	hreg->irq = pHCB->HCS_Intr;
	hreg->this_id = pHCB->HCS_SCSI_ID;	/* Assign HCS index           */
// 02/20/2001
#if LINUX_VERSION_CODE >= CVT_LINUX_VERSION(2,3,99)
	hreg->base = (unsigned long) pHCB;
	hreg2 = hreg;
#if CONFIG_HOTPLUG
	cb32_dev_id = hreg2;
#endif
#else
	hreg->base = (UCHAR *) pHCB;
	hreg2 = NULL;
#endif

#if 1
	hreg->sg_tablesize = TOTAL_SG_ENTRY;	/* Maximun support is 32 */
#else
	hreg->sg_tablesize = SG_NONE;	/* No SG                        */
#endif

	/* Initial orc chip           */
	switch (i) {
#if LINUX_VERSION_CODE >= CVT_LINUX_VERSION(1,3,0)
	case 0:
	    ok = request_irq(pHCB->HCS_Intr, cb32_intr0, SA_INTERRUPT | SA_SHIRQ, "cb32", hreg2);
	    break;
	case 1:
	    ok = request_irq(pHCB->HCS_Intr, cb32_intr1, SA_INTERRUPT | SA_SHIRQ, "cb32", hreg2);
	    break;
	case 2:
	    ok = request_irq(pHCB->HCS_Intr, cb32_intr2, SA_INTERRUPT | SA_SHIRQ, "cb32", hreg2);
	    break;
	case 3:
	    ok = request_irq(pHCB->HCS_Intr, cb32_intr3, SA_INTERRUPT | SA_SHIRQ, "cb32", hreg2);
	    break;
	case 4:
	    ok = request_irq(pHCB->HCS_Intr, cb32_intr4, SA_INTERRUPT | SA_SHIRQ, "cb32", hreg2);
	    break;
	case 5:
	    ok = request_irq(pHCB->HCS_Intr, cb32_intr5, SA_INTERRUPT | SA_SHIRQ, "cb32", hreg2);
	    break;
	case 6:
	    ok = request_irq(pHCB->HCS_Intr, cb32_intr6, SA_INTERRUPT | SA_SHIRQ, "cb32", hreg2);
	    break;
	case 7:
	    ok = request_irq(pHCB->HCS_Intr, cb32_intr7, SA_INTERRUPT | SA_SHIRQ, "cb32", hreg2);
	    break;
	default:
	    cb32_panic("cb32_cb: Too many host adapters\n");
	    break;
	}

	if (ok < 0) {
	    if (ok == -EINVAL) {
		printk("cb32_cb: bad IRQ %d.\n", pHCB->HCS_Intr);
		printk("         Contact author.\n");
	    } else {
		if (ok == -EBUSY)
	  	    printk("cb32_cb: IRQ %d already in use. Configure another.\n", pHCB->HCS_Intr);
		else {
		    printk("\ncb32_cb: Unexpected error code on requesting IRQ %d.\n", pHCB->HCS_Intr);
		    printk("         Contact author.\n");
		}
	    }
	    cb32_panic("cb32_cb: driver needs an IRQ.\n");
	}
#endif
    }

    tpnt->this_id = -1;
    tpnt->can_queue = 1;
    return 1;
}

#if CONFIG_HOTPLUG
/*****************************************************************************
 Function name  : cb32_release
 Description    : 
 Input          : pHCB  -       Pointer to host adapter structure
 Output         : None.
 Return         : pSRB  -       Pointer to SCSI request block.
*****************************************************************************/
int cb32_release(struct Scsi_Host * shpnt)
{

    free_irq(shpnt->irq, cb32_dev_id);
    release_region(shpnt->io_port, shpnt->n_io_port);
    return 0;
}
#endif

/*****************************************************************************
 Function name  : cb32BuildSCB
 Description    : 
 Input          : pHCB  -       Pointer to host adapter structure
 Output         : None.
 Return         : pSRB  -       Pointer to SCSI request block.
*****************************************************************************/
static void cb32BuildSCB(ORC_HCS * pHCB, ORC_SCB * pSCB, Scsi_Cmnd * SCpnt)
{				/* Create corresponding SCB     */
	struct scatterlist *pSrbSG;
	ORC_SG *pSG;		/* Pointer to SG list           */
	int i;
	U32 TotalLen;
	ESCB *pEScb;

	pEScb = pSCB->SCB_EScb;
	pEScb->SCB_Srb = SCpnt;
	pSG = NULL;

	pSCB->SCB_Opcode = ORC_EXECSCSI;
	pSCB->SCB_Flags = SCF_NO_DCHK;	/* Clear done bit               */
	pSCB->SCB_Target = SCpnt->target;
	pSCB->SCB_Lun = SCpnt->lun;
	pSCB->SCB_Reserved0 = 0;
	pSCB->SCB_Reserved1 = 0;
	pSCB->SCB_SGLen = 0;

	if ((pSCB->SCB_XferLen = (U32) SCpnt->request_bufflen)) {
		pSG = (ORC_SG *) & pEScb->ESCB_SGList[0];
		if (SCpnt->use_sg) {
			TotalLen = 0;
			pSCB->SCB_SGLen = (U32) (SCpnt->use_sg * 8);
			pSrbSG = (struct scatterlist *) SCpnt->request_buffer;
			for (i = 0; i < SCpnt->use_sg; i++, pSG++, pSrbSG++) {
#if LINUX_VERSION_CODE >= CVT_LINUX_VERSION(1,3,0)
				pSG->SG_Ptr = (U32) (VIRT_TO_BUS(pSrbSG->address));
#else
				pSG->SG_Ptr = (U32) pSrbSG->address;
#endif
				pSG->SG_Len = (U32) pSrbSG->length;
				TotalLen += (U32) pSrbSG->length;
			}
		} else {	/* Non SG                       */
			pSCB->SCB_SGLen = 0x8;
#if LINUX_VERSION_CODE >= CVT_LINUX_VERSION(1,3,0)
			pSG->SG_Ptr = (U32) (VIRT_TO_BUS(SCpnt->request_buffer));
#else
			pSG->SG_PTR = (U32) SCpnt->request_buffer;
#endif
			pSG->SG_Len = (U32) SCpnt->request_bufflen;
		}
	}
	pSCB->SCB_SGPAddr = (U32) pSCB->SCB_SensePAddr;
	pSCB->SCB_HaStat = 0;
	pSCB->SCB_TaStat = 0;
	pSCB->SCB_Link = 0xFF;
	pSCB->SCB_SenseLen = SENSE_SIZE;
	pSCB->SCB_CDBLen = SCpnt->cmd_len;
	if (pSCB->SCB_CDBLen >= IMAX_CDB) {
		printk("max cdb length= %x\b", SCpnt->cmd_len);
		pSCB->SCB_CDBLen = IMAX_CDB;
	}
	pSCB->SCB_Ident = SCpnt->lun | DISC_ALLOW;
	if (SCpnt->device->tagged_supported) {	/* Tag Support                  */
		pSCB->SCB_TagMsg = SIMPLE_QUEUE_TAG;	/* Do simple tag only   */
	} else {
		pSCB->SCB_TagMsg = 0;	/* No tag support               */
	}
	memcpy(&pSCB->SCB_CDB[0], &SCpnt->cmnd, pSCB->SCB_CDBLen);
	return;
}

/*****************************************************************************
 Function name  : cb32_queue
 Description    : Queue a command and setup interrupts for a free bus.
 Input          : pHCB  -       Pointer to host adapter structure
 Output         : None.
 Return         : pSRB  -       Pointer to SCSI request block.
*****************************************************************************/
int cb32_queue(Scsi_Cmnd * SCpnt, void (*done) (Scsi_Cmnd *))
{
	register ORC_SCB *pSCB;
	ORC_HCS *pHCB;		/* Point to Host adapter control block */

	if (SCpnt->lun > 16) {
		SCpnt->result = (DID_TIME_OUT << 16);
		done(SCpnt);	/* Notify system DONE           */
		return (0);
	}
	pHCB = (ORC_HCS *) SCpnt->host->base;
	SCpnt->scsi_done = done;
	/* Get free SCSI control block  */
	if ((pSCB = orc_alloc_scb(pHCB)) == NULL) {
		cb32AppendSRBToQueue(pHCB, SCpnt);	/* Buffer this request  */
		/* printk("cb32_entry: can't allocate SCB\n"); */
		return (0);
	}
	cb32BuildSCB(pHCB, pSCB, SCpnt);
	orc_exec_scb(pHCB, pSCB);	/* Start execute SCB            */

	return (0);
}

/*****************************************************************************
 Function name  : cb32_command
 Description    : We only support command in interrupt-driven fashion
 Input          : pHCB  -       Pointer to host adapter structure
 Output         : None.
 Return         : pSRB  -       Pointer to SCSI request block.
*****************************************************************************/
int cb32_command(Scsi_Cmnd * SCpnt)
{
	printk("cb32_cb: interrupt driven driver; use cb32_queue()\n");
	return -1;
}

/*****************************************************************************
 Function name  : cb32_abort
 Description    : Abort a queued command.
	                 (commands that are on the bus can't be aborted easily)
 Input          : pHCB  -       Pointer to host adapter structure
 Output         : None.
 Return         : pSRB  -       Pointer to SCSI request block.
*****************************************************************************/
int cb32_abort(Scsi_Cmnd * SCpnt)
{
	ORC_HCS *hcsp;
	int	result;	// 2001-05-21

	hcsp = (ORC_HCS *) SCpnt->host->base;
	// 2001-05-21 add to following
	result = orc_abort_srb(hcsp, (ULONG) SCpnt);
	if (SCpnt->scsi_done == NULL) {
		printk("cb32_abort: Scsi_Cmnd done == NULL \n");
		return result;
	}
	if ((result == SCSI_ABORT_SUCCESS)
	    || (result == SCSI_ABORT_NOT_RUNNING))
		SCpnt->scsi_done(SCpnt);
	return result;
	// 2001-05-21 return orc_abort_srb(hcsp, (ULONG) SCpnt);
}

/*****************************************************************************
 Function name  : cb32_reset
 Description    : Reset registers, reset a hanging bus and
                  kill active and disconnected commands for target w/o soft reset
 Input          : pHCB  -       Pointer to host adapter structure
 Output         : None.
 Return         : pSRB  -       Pointer to SCSI request block.
*****************************************************************************/
int cb32_reset(Scsi_Cmnd * SCpnt, unsigned int reset_flags)
{				/* I need Host Control Block Information */
	ORC_HCS *pHCB;
	pHCB = (ORC_HCS *) SCpnt->host->base;

	if (reset_flags & (SCSI_RESET_SUGGEST_BUS_RESET | SCSI_RESET_SUGGEST_HOST_RESET))
		return orc_reset_scsi_bus(pHCB);
	else
		return orc_device_reset(pHCB, (ULONG) SCpnt, SCpnt->target, reset_flags);

}

/*****************************************************************************
 Function name  : cb32SCBPost
 Description    : This is callback routine be called when orc finish one
			SCSI command.
 Input          : pHCB  -       Pointer to host adapter control block.
		  pSCB  -       Pointer to SCSI control block.
 Output         : None.
 Return         : None.
*****************************************************************************/
void cb32SCBPost(BYTE * pHcb, BYTE * pScb)
{
    Scsi_Cmnd *pSRB;	/* Pointer to SCSI request block */
    ORC_HCS *pHCB;
    ORC_SCB *pSCB;
    ESCB *pEScb;

    pHCB = (ORC_HCS *) pHcb;
    pSCB = (ORC_SCB *) pScb;
    pEScb = pSCB->SCB_EScb;
    if ((pSRB = (Scsi_Cmnd *) pEScb->SCB_Srb) == 0) {
	printk("cb32SCBPost: SRB pointer is empty\n");
	orc_release_scb(pHCB, pSCB);	/* Release SCB for current channel */
		return;
    }
    pEScb->SCB_Srb = NULL;

    switch (pSCB->SCB_HaStat) {
    case 0x0:
    case 0xa:	/* Linked command complete without error and linked normally */
    case 0xb:	/* Linked command complete without error interrupt generated */
	pSCB->SCB_HaStat = 0;
	break;

    case 0x11:	/* Selection time out-The initiator selection or target
		   reselection was not complete within the SCSI Time out period */
	pSCB->SCB_HaStat = DID_TIME_OUT;
	break;

    case 0x14:	/* Target bus phase sequence failure-An invalid bus phase or bus
		   phase sequence was requested by the target. The host adapter
		   will generate a SCSI Reset Condition, notifying the host with
		   a SCRD interrupt */
	pSCB->SCB_HaStat = DID_RESET;
	break;

    case 0x1a:	/* SCB Aborted. 07/21/98 */
	pSCB->SCB_HaStat = DID_ABORT;
	break;

    case 0x12:	/* Data overrun/underrun-The target attempted to transfer more data
		   than was allocated by the Data Length field or the sum of the
		   Scatter / Gather Data Length fields. */
    case 0x13:	/* Unexpected bus free-The target dropped the SCSI BSY at an unexpected time. */
    case 0x16:	/* Invalid CCB Operation Code-The first byte of the CCB was invalid. */

    default:
	printk("cb32: %x %x\n", pSCB->SCB_HaStat, pSCB->SCB_TaStat);
	pSCB->SCB_HaStat = DID_ERROR;	/* Couldn't find any better */
	break;
    }

    if (pSCB->SCB_TaStat == 2) {	/* Check condition              */
	memcpy((unsigned char *) &pSRB->sense_buffer[0],
	   (unsigned char *) &pEScb->ESCB_SGList[0], SENSE_SIZE);
    }
    pSRB->result = pSCB->SCB_TaStat | (pSCB->SCB_HaStat << 16);
    pSRB->scsi_done(pSRB);	/* Notify system DONE           */

	/* Find the next pending SRB    */
    if ((pSRB = cb32PopSRBFromQueue(pHCB)) != NULL) {	/* Assume resend will success   */
	/* Reuse old SCB                */
	cb32BuildSCB(pHCB, pSCB, pSRB);	/* Create corresponding SCB     */
	orc_exec_scb(pHCB, pSCB);	/* Start execute SCB            */
    } else {		/* No Pending SRB               */
	orc_release_scb(pHCB, pSCB);	/* Release SCB for current channel */
    }
    return;
}

/*****************************************************************************
 Function name  : cb32_biosparam
 Description    : Return the "logical geometry"
 Input          : pHCB  -       Pointer to host adapter structure
 Output         : None.
 Return         : pSRB  -       Pointer to SCSI request block.
*****************************************************************************/
#if LINUX_VERSION_CODE >= CVT_LINUX_VERSION(1,3,0)
int cb32_biosparam(Scsi_Disk * disk, kdev_t dev, int *info_array)
#else
int cb32_biosparam(Scsi_Disk * disk, int dev, int *info_array)
#endif
{
	ORC_HCS *pHcb;		/* Point to Host adapter control block */
	ORC_TCS *pTcb;

	pHcb = (ORC_HCS *) disk->device->host->base;
	pTcb = &pHcb->HCS_Tcs[disk->device->id];

	if (pTcb->TCS_DrvHead) {
		info_array[0] = pTcb->TCS_DrvHead;
		info_array[1] = pTcb->TCS_DrvSector;
		info_array[2] = disk->capacity / pTcb->TCS_DrvHead / pTcb->TCS_DrvSector;
	} else {
		if (pTcb->TCS_DrvFlags & TCF_DRV_255_63) {
			info_array[0] = 255;
			info_array[1] = 63;
			info_array[2] = disk->capacity / 255 / 63;
		} else {
			info_array[0] = 64;
			info_array[1] = 32;
			info_array[2] = disk->capacity >> 11;
		}
	}
	return 0;
}


static void subIntr(ORC_HCS * pHCB, int irqno)
{
#if LINUX_VERSION_CODE >= CVT_LINUX_VERSION(2,1,95)
	unsigned long flags;

	spin_lock_irqsave(&io_request_lock, flags);
#endif

	if (pHCB->HCS_Intr != irqno) {
#if LINUX_VERSION_CODE >= CVT_LINUX_VERSION(2,1,95)
		spin_unlock_irqrestore(&io_request_lock, flags);
#endif
		return;
	}
	orc_interrupt(pHCB);

#if LINUX_VERSION_CODE >= CVT_LINUX_VERSION(2,1,95)
	spin_unlock_irqrestore(&io_request_lock, flags);
#endif
}

/*
 * Interrupts handler (main routine of the driver)
 */
#if LINUX_VERSION_CODE >= CVT_LINUX_VERSION(1,3,0)
static void cb32_intr0(int irqno, void *dev_id, struct pt_regs *regs)
#else
static void cb32_intr0(int irqno, struct pt_regs *regs)
#endif
{
	subIntr(&orc_hcs[0], irqno);
}

#if LINUX_VERSION_CODE >= CVT_LINUX_VERSION(1,3,0)
static void cb32_intr1(int irqno, void *dev_id, struct pt_regs *regs)
#else
static void cb32_intr1(int irqno, struct pt_regs *regs)
#endif
{
	subIntr(&orc_hcs[1], irqno);
}

#if LINUX_VERSION_CODE >= CVT_LINUX_VERSION(1,3,0)
static void cb32_intr2(int irqno, void *dev_id, struct pt_regs *regs)
#else
static void cb32_intr2(int irqno, struct pt_regs *regs)
#endif
{
	subIntr(&orc_hcs[2], irqno);
}

#if LINUX_VERSION_CODE >= CVT_LINUX_VERSION(1,3,0)
static void cb32_intr3(int irqno, void *dev_id, struct pt_regs *regs)
#else
static void cb32_intr3(int irqno, struct pt_regs *regs)
#endif
{
	subIntr(&orc_hcs[3], irqno);
}

#if LINUX_VERSION_CODE >= CVT_LINUX_VERSION(1,3,0)
static void cb32_intr4(int irqno, void *dev_id, struct pt_regs *regs)
#else
static void cb32_intr4(int irqno, struct pt_regs *regs)
#endif
{
	subIntr(&orc_hcs[4], irqno);
}

#if LINUX_VERSION_CODE >= CVT_LINUX_VERSION(1,3,0)
static void cb32_intr5(int irqno, void *dev_id, struct pt_regs *regs)
#else
static void cb32_intr5(int irqno, struct pt_regs *regs)
#endif
{
	subIntr(&orc_hcs[5], irqno);
}

#if LINUX_VERSION_CODE >= CVT_LINUX_VERSION(1,3,0)
static void cb32_intr6(int irqno, void *dev_id, struct pt_regs *regs)
#else
static void cb32_intr6(int irqno, struct pt_regs *regs)
#endif
{
	subIntr(&orc_hcs[6], irqno);
}

#if LINUX_VERSION_CODE >= CVT_LINUX_VERSION(1,3,0)
static void cb32_intr7(int irqno, void *dev_id, struct pt_regs *regs)
#else
static void cb32_intr7(int irqno, struct pt_regs *regs)
#endif
{
	subIntr(&orc_hcs[7], irqno);
}

/* 
 * Dump the current driver status and panic...
 */
static void cb32_panic(char *msg)
{
	printk("\ncb32_panic: %s\n", msg);
	panic("cb32 panic");
}
/*===================================================================*/

static dev_node_t *cb32_config(dev_locator_t *loc)
{
    Scsi_Device *dev;
    dev_node_t  *node;
    int n = 0;
#if LINUX_VERSION_CODE >= CVT_LINUX_VERSION(2,1,75)
    struct Scsi_Host *host;
#endif


   node = kmalloc(7 * sizeof(dev_node_t), GFP_KERNEL);
#if LINUX_VERSION_CODE < CVT_LINUX_VERSION(2,1,75)
    for (dev = scsi_devices; dev != NULL; dev = dev->next) {
	if (dev->host->hostt == &driver_template) {
	    { // Dummy
#else
    for (host = scsi_hostlist; host; host = host->next) {
	if (host->hostt == &driver_template) {
	    for (dev = host->host_queue; dev; dev = dev->next) {
#endif
		u_long arg[2], id;
		kernel_scsi_ioctl(dev, SCSI_IOCTL_GET_IDLUN, arg);
		id = (arg[0]&0x0f) + ((arg[0]>>4)&0xf0) +
		    ((arg[0]>>8)&0xf00) + ((arg[0]>>4)&0xf000);
    printk(KERN_INFO "cb32_cb : kernel_scsi_ioctl id = %ld\n", id);
		node->minor = 0;
		switch (dev->type) {
		case TYPE_TAPE:
		    node->major = SCSI_TAPE_MAJOR;
		    sprintf(node->dev_name, "st#%04lx", id);
		    break;
		case TYPE_DISK:
		case TYPE_MOD:
		    node->major = SCSI_DISK0_MAJOR;
		    sprintf(node->dev_name, "sd#$04lx", id);
		    break;
		case TYPE_ROM:
		case TYPE_WORM:
		    node->major = SCSI_CDROM_MAJOR;
		    sprintf(node->dev_name, "sr#$04lx", id);
		    break;
		default:
		    node->major = SCSI_GENERIC_MAJOR;
		    sprintf(node->dev_name, "sg#$04lx", id);
		    break;
		} // switch (dev->type)
                if (n) node[n-1].next = &node[n];
		n++;

	    } // for (dev = ... OR dummy
	} // if
    } // for

    if (n == 0) {
	printk(KERN_INFO "cb32_cb: no SCSI devices found\n");
	kfree(node);
	node = NULL;
    } else {
	printk(KERN_INFO "cb32_cb: %d SCSI devices found\n", n);
	node[n-1].next = NULL;
    }

    return node;
}

/*===================================================================*/

static dev_node_t *cb32_attach(dev_locator_t *loc)
{
    u_int  io;
    u_char bus, devfn, irq;
    dev_node_t *node;
    char s[20];

    if (loc->bus != LOC_PCI) {
	printk(KERN_INFO "cb32_cb : not CardBus (PCI) \n");
	return NULL;
    }

    bus = loc->b.pci.bus; devfn = loc->b.pci.devfn;
    printk(KERN_INFO "cb32_cb : attach(bus %d, fn %d)\n", bus, devfn);
    pcibios_read_config_dword(bus, devfn, PCI_BASE_ADDRESS_0, &io);
    pcibios_read_config_byte(bus, devfn, PCI_INTERRUPT_LINE, &irq);
    printk(KERN_INFO "cb32_cb : io = 0x%04X irq = %d\n", io, irq);

    sprintf(s, "cb32_setup");
    cb32_setup(s, ints);
    printk(KERN_INFO "cb32_cb : cb32_setup\n");
    scsi_register_module(MODULE_SCSI_HA, &driver_template);
    if (! driver_template.present) {
	scsi_unregister_module(MODULE_SCSI_HA, &driver_template);
	return NULL;
    }

    printk(KERN_INFO "cb32_cb : scsi_register_module\n");

    node = cb32_config(loc);
    if (node != NULL)
	MOD_INC_USE_COUNT;

    return node;
}

static void cb32_detach(dev_node_t *node)
{
    printk(KERN_INFO "cb32_cb : cb32_detach\n");
    DEBUG(0, "cb32_detach \n");
    scsi_unregister_module(MODULE_SCSI_HA, &driver_template);
    printk(KERN_INFO "cb32_cb : scsi_unregister_module\n");
    MOD_DEC_USE_COUNT;
    if (node) kfree(node);
    return;
}

#if CONFIG_HOTPLUG
static int __devinit cb32_probe(struct pci_dev *pdev, const struct pci_device_id *id)
{
    unsigned char  chip_rev;
    unsigned short tmp16;
    char s[20];
    int base_addr;
    dev_node_t *node;

      if (pci_enable_device(pdev))
              return -ENODEV;

      /* disabe all powermanagement */
      pci_write_config_dword(pdev, PCI_POWERMGMT, 0x0000);

      pci_set_master(pdev);

      pci_read_config_word(pdev, PCI_STATUS, &tmp16);
      pci_write_config_word(pdev, PCI_STATUS, &tmp16);

      pci_read_config_byte(pdev, PCI_REVISION_ID, &chip_rev);

      base_addr = pci_resource_start(pdev, 0);

      if (!request_region(base_addr, 256, "cb32_cb")) {
        printk(KERN_ERR "cb32_probe: failed to allocate io-region\n");
        return -ENODEV;
      }

      sprintf(s, "cb32_setup");
      cb32_setup(s, ints);

      scsi_register_module(MODULE_SCSI_HA, &driver_template);
      if (! driver_template.present) {
        scsi_unregister_module(MODULE_SCSI_HA, &driver_template);
        release_region(base_addr, 256);
        return -ENODEV;
      }

      node = cb32_config(NULL);
      if (node == NULL) {
        scsi_unregister_module(MODULE_SCSI_HA, &driver_template);
        release_region(base_addr, 256);
        return -ENODEV;
      }

      MOD_INC_USE_COUNT;
      return 0;  
}

static void __devexit cb32_remove(struct pci_dev *pdev)
{
    int base_addr;

//  printk("cb32_cb: cb32_remove\n");
    base_addr = pci_resource_start(pdev, 0);

    scsi_unregister_module(MODULE_SCSI_HA, &driver_template);
    MOD_DEC_USE_COUNT;
    printk("cb32_cb: cb32_remove : finish\n");
    return;
} 

#endif

/*===================================================================*/

#if LINUX_VERSION_CODE < CVT_LINUX_VERSION(2,4,0)
int init_module(void)
#else
static int __init init_cb32_cb(void)
#endif
{
    printk(KERN_INFO "cb32_cb : %s start\n", cb32_Version);

#if CONFIG_HOTPLUG
    pci_register_driver(&cb32_pci_ops);
#else
    register_driver(&cb32_ops);
#endif
    return 0;
}

#if LINUX_VERSION_CODE < CVT_LINUX_VERSION(2,4,0)
void cleanup_module(void)
#else
static void __exit exit_cb32_cb(void)
#endif
{
    printk(KERN_INFO "cb32_cb : quit\n");
#if CONFIG_HOTPLUG
    pci_unregister_driver(&cb32_pci_ops);
#else
    unregister_driver(&cb32_ops);
#endif
}
#if LINUX_VERSION_CODE >= CVT_LINUX_VERSION(2,4,0)
module_init(init_cb32_cb);
module_exit(exit_cb32_cb);
#endif
/*#include "cb32scsi.c" */
