/**************************************************************************
 * CB32 (Inic 850) device driver for Linux.
 *
 * Copyright (c) 1994-1998 Initio Corporation
 * Copyright (c) 2000 RATOC Systems, Inc.
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * --------------------------------------------------------------------------
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions, and the following disclaimer,
 *    without modification, immediately at the beginning of the file.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * Where this Software is combined with software released under the terms of 
 * the GNU Public License ("GPL") and the terms of the GPL would require the 
 * combined work to also be released under the terms of the GPL, the terms
 * and conditions of this License will apply in addition to those of the
 * GPL with the exception of any terms or conditions of this License that
 * conflict with, or are expressly prohibited by, the GPL.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *************************************************************************
 *
 * module: cb32scsi.c 
 * DESCRIPTION:
 * 	This is the Linux low-level SCSI driver for RATOC CB32 SCSI host
 * adapters
 *
 * 07/02/98 hl	- v.91n Initial drivers.
 * 09/14/98 hl - v1.01 Support new Kernel.
 * 09/22/98 hl - v1.01a Support reset.
 * 09/24/98 hl - v1.01b Fixed reset.
 * 10/05/98 hl - v1.02 split the source code and release.
 * 12/19/98 bv - v1.02a Use spinlocks for 2.1.95 and up
 * 05/10/00 ko - v1.00 Inic 850 new
 * 07/10/00 sk - v1.00 CB32
 **************************************************************************/

#ifndef CVT_LINUX_VERSION
#define CVT_LINUX_VERSION(V,P,S)        (V * 65536 + P * 256 + S)
#endif

#ifdef DEBUG
#define DBG
#endif

#include <linux/sched.h>
#include <asm/io.h>
#include <asm/delay.h>
#include "cb32scsi.h"


int cb32_load_FW(ORC_HCS * hcsp);
/* ---- INTERNAL FUNCTIONS ---- */
static UCHAR waitChipReady(ORC_HCS * hcsp);
static UCHAR waitFWReady(ORC_HCS * hcsp);
static UCHAR waitFWReady(ORC_HCS * hcsp);
static UCHAR waitSCSIRSTdone(ORC_HCS * hcsp);
static UCHAR waitHDOoff(ORC_HCS * hcsp);
static UCHAR waitHDIset(ORC_HCS * hcsp, UCHAR * pData);
static unsigned short get_FW_version(ORC_HCS * hcsp);
static UCHAR set_NVRAM(ORC_HCS * hcsp, unsigned char address, unsigned char value);
static UCHAR get_NVRAM(ORC_HCS * hcsp, unsigned char address, unsigned char *pDataIn);
static int se2_rd_all(ORC_HCS * hcsp);
static void se2_update_all(ORC_HCS * hcsp);	/* setup default pattern        */
static void read_eeprom(ORC_HCS * hcsp);
static UCHAR load_FW(ORC_HCS * hcsp);
static void setup_SCBs(ORC_HCS * hcsp);
static void initAFlag(ORC_HCS * hcsp);
ORC_SCB *orc_alloc_scb(ORC_HCS * hcsp);

/* ---- EXTERNAL FUNCTIONS ---- */
extern void cb32SCBPost(BYTE * pHcb, BYTE * pScb);

#if LINUX_VERSION_CODE < CVT_LINUX_VERSION(2,1,95)
#define spin_lock_irqsave(p,f)	{save_flags(f); cli();}
#define spin_unlock_irqrestore(p,f)    restore_flags(f)
#endif

/* ---- INTERNAL VARIABLES ---- */
ORC_HCS orc_hcs[MAX_SUPPORTED_ADAPTERS];
static CB32_ADPT_STRUCT cb32_adpt[MAX_SUPPORTED_ADAPTERS];
/* set by cb32_setup according to the command line */
int orc_num_scb;

NVRAM nvram, *nvramp = &nvram;

static UCHAR dftNvRam[NVRAM_SIZE] =
{
/*----------for CardBus-------------*/
    0x95,		/* 0x00: Sub System Vendor ID 0 */
    0x11,		/* 0x01: Sub System Vendor ID 1 */
    0x51,		/* 0x02: Sub System ID 0  851:CB32 853:PCI31/33 */
    0x08,		/* 0x03: Sub System ID 1        */
    0x00,		/* 0x04: SubClass               */
    0x95,		/* 0x05: Vendor ID 0            */
    0x11,		/* 0x06: Vendor ID 1            */
    0x51,		/* 0x07: Device ID 0            */
    0x08,		/* 0x08: Device ID 1            */
    0x3F,0x00,0x00,0x00,	/* 0x09-0x0C:           */
    0xFF,0xFF,0xFF,		/* 0x0D-0x0F: Reserved  */
    0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,	/* 0x10-0x17 */
    0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,	/* 0x18-0x1F */
    0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,	/* 0x20-0x27 */
    0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,	/* 0x28-0x2F */
    0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,	/* 0x30-0x37 */
    0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,	/* 0x38-0x3F */
    0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,	/* 0x40-0x47 */
    0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,	/* 0x48-0x4F */
    0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,	/* 0x50-0x57 */
    0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,	/* 0x58-0x5F */
    0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,	/* 0x60-0x67 */
    0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,	/* 0x68-0x6F */
    0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,	/* 0x70-0x77 */
    0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,	/* 0x78-0x7F */
    0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,	/* 0x80-0x87 */
    0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,	/* 0x88-0x8F */
    0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,	/* 0x90-0x97 */
    0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,	/* 0x98-0x9F */
    0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,	/* 0xA0-0xA7 */
    0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,	/* 0xA8-0xAF */
    0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,	/* 0xB0-0xA7 */
    0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,	/* 0xB8-0xAF */

/*----------header (Start 0xC0)-------------*/
	0x95,		/* 0x00: Sub System Vendor ID 0 */
	0x11,		/* 0x01: Sub System Vendor ID 1 */
	0x51,		/* 0x02: Sub System ID 0  851:CB32 853:PCI31/33 */
	0x08,		/* 0x03: Sub System ID 1        */
	0x00,		/* 0x04: SubClass               */
	0x95,		/* 0x05: Vendor ID 0            */
	0x11,		/* 0x06: Vendor ID 1            */
	0x51,		/* 0x07: Device ID 0            */
	0x08,		/* 0x08: Device ID 1            */
	0x00,		/* 0x09: Reserved               */
	0x00,		/* 0x0A: Reserved               */
	0x01,		/* 0x0B: Revision of Data Structure     */
		/* -- Host Adapter Structure --- */
	0x01,		/* 0x0C: Number Of SCSI Channel */
	0x01,		/* 0x0D: BIOS Configuration 1   */
	0x00,		/* 0x0E: BIOS Configuration 2   */
	0x00,		/* 0x0F: BIOS Configuration 3   */
		/* --- SCSI Channel 0 Configuration --- */
	0x07,		/* 0x10: H/A ID                 */
	0x83,		/* 0x11: Channel Configuration  */
	0x20,		/* 0x12: MAX TAG per target     */
	0x05,		/* 0x13: SCSI Reset Recovering time     */
	0x10,		/* 0x14: Channel Configuration4 */
	0x05,		/* 0x15: Channel Configuration5 */
		/* SCSI Channel 0 Target Configuration  */
			/* 0x16-0x25                    */
	0xC9, 0xC9, 0xC9, 0xC9, 0xC9, 0xC9, 0xC9, 0xC9,
	0xC9, 0xC9, 0xC9, 0xC9, 0xC9, 0xC9, 0xC9, 0xC9,
		/* --- SCSI Channel 1 Configuration --- */
	0x07,		/* 0x26: H/A ID                 */
	0x83,		/* 0x27: Channel Configuration  */
	0x20,		/* 0x28: MAX TAG per target     */
	0x0A,		/* 0x29: SCSI Reset Recovering time     */
	0x00,		/* 0x2A: Channel Configuration4 */
	0x00,		/* 0x2B: Channel Configuration5 */
		/* SCSI Channel 1 Target Configuration  */
			/* 0x2C-0x3B                    */
	0xC9, 0xC9, 0xC9, 0xC9, 0xC9, 0xC9, 0xC9, 0xC9,
	0xC9, 0xC9, 0xC9, 0xC9, 0xC9, 0xC9, 0xC9, 0xC9,
	0x00,		/* 0x3C: Reserved               */
	0x00,		/* 0x3D: Reserved               */
	0x00,		/* 0x3E: Reserved               */
	0x00		/* 0x3F: Checksum               */
};


/***************************************************************************
Function name	: waitForPause
***************************************************************************/
static void
waitForPause(unsigned amount)
/*
 Arguments	: unsgined amount;	milisecond
*/
{
    int i;

    for (i = 0; i < amount; i++)
	udelay(1000);	/* 1micro x 1000 = 1mili */
}

/***************************************************************************/
UCHAR waitChipReady(ORC_HCS * hcsp)
{
    int i;

    for (i = 0; i < 2000; i++) {	/* Wait 1 second for report timeout */
	if (ORC_RD(hcsp->HCS_Base, ORC_HCTRL) & HOSTSTOP)	/* Wait HOSTSTOP set */
	    return (TRUE);
	waitForPause(5);	/* wait 500ms before try again  */
    }
    return (FALSE);
}

/***************************************************************************/
UCHAR waitFWReady(ORC_HCS * hcsp)
{
	int i;

	/* Wait 1 second for report timeout     */
	for (i = 0; i < 2000; i++) {
		/* Wait READY set */
		if (ORC_RD(hcsp->HCS_Base, ORC_HSTUS) & RREADY)	
			return (TRUE);
		waitForPause(5);	/* wait 500ms before try again  */
	}
	return (FALSE);
}

/***************************************************************************/
UCHAR waitSCSIRSTdone(ORC_HCS * hcsp)
{
    int i;

#ifdef DBG
	printk("cb32_cb: waitSCSIRSTdone");
#endif
    for (i = 0; i < 2000; i++) { /* Wait 1 second for report timeout */
 	/* Wait SCSIRST done */
	if (!(ORC_RD(hcsp->HCS_Base, ORC_HCTRL) & SCSIRST)) {
#ifdef DBG
	printk(" reset ok\n");
#endif
	    return (TRUE);
	}
	waitForPause(5);	/* wait 500ms before try again  */
    }
#ifdef DBG
	printk(" reset time-out\n");
#endif
    return (FALSE);
}

/***************************************************************************/
UCHAR waitHDOoff(ORC_HCS * hcsp)
{
    int i;

    for (i = 0; i < 2000; i++) { /* Wait 1 second for report timeout */
	/* Wait HDO off */
	if (!(ORC_RD(hcsp->HCS_Base, ORC_HCTRL) & HDO))
	    return (TRUE);
	waitForPause(5);	/* wait 500ms before try again  */
    }
    return (FALSE);
}

/***************************************************************************/
UCHAR waitHDIset(ORC_HCS * hcsp, UCHAR * pData)
{
    int i;

    for (i = 0; i < 2000; i++) {	/* Wait 1 second for report timeout */
	if ((*pData = ORC_RD(hcsp->HCS_Base, ORC_HSTUS)) & HDI)
	    return (TRUE);	/* Wait HDI set */
	waitForPause(5);	/* wait 500ms before try again  */
    }
    return (FALSE);
}

/***************************************************************************/
unsigned short get_FW_version(ORC_HCS * hcsp)
{
    UCHAR bData;
    union {
	unsigned short sVersion;
	unsigned char cVersion[2];
    } Version;

    ORC_WR(hcsp->HCS_Base + ORC_HDATA, ORC_CMD_VERSION);
    ORC_WR(hcsp->HCS_Base + ORC_HCTRL, HDO);
    if (waitHDOoff(hcsp) == FALSE)	/* Wait HDO off   */
	return (FALSE);

    if (waitHDIset(hcsp, &bData) == FALSE)	/* Wait HDI set   */
	return (FALSE);
    Version.cVersion[0] = ORC_RD(hcsp->HCS_Base, ORC_HDATA);
    ORC_WR(hcsp->HCS_Base + ORC_HSTUS, bData);	/* Clear HDI      */

    if (waitHDIset(hcsp, &bData) == FALSE)	/* Wait HDI set   */
	return (FALSE);
    Version.cVersion[1] = ORC_RD(hcsp->HCS_Base, ORC_HDATA);
    ORC_WR(hcsp->HCS_Base + ORC_HSTUS, bData);	/* Clear HDI      */

    return (Version.sVersion);
}

/***************************************************************************/
UCHAR set_NVRAM(ORC_HCS * hcsp, unsigned char address, unsigned char value)
{
    ORC_WR(hcsp->HCS_Base + ORC_HDATA, ORC_CMD_SET_NVM);  /* Write command */
    ORC_WR(hcsp->HCS_Base + ORC_HCTRL, HDO);
    if (waitHDOoff(hcsp) == FALSE)	/* Wait HDO off   */
	return (FALSE);

    ORC_WR(hcsp->HCS_Base + ORC_HDATA, address);	/* Write address */
    ORC_WR(hcsp->HCS_Base + ORC_HCTRL, HDO);
    if (waitHDOoff(hcsp) == FALSE)	/* Wait HDO off   */
	return (FALSE);

    ORC_WR(hcsp->HCS_Base + ORC_HDATA, value);	/* Write value  */
    ORC_WR(hcsp->HCS_Base + ORC_HCTRL, HDO);
    if (waitHDOoff(hcsp) == FALSE)	/* Wait HDO off   */
	return (FALSE);

    return (TRUE);
}

/***************************************************************************/
UCHAR get_NVRAM(ORC_HCS * hcsp, unsigned char address, unsigned char *pDataIn)
{
    unsigned char bData;

    ORC_WR(hcsp->HCS_Base + ORC_HDATA, ORC_CMD_GET_NVM);  /* Write command */
    ORC_WR(hcsp->HCS_Base + ORC_HCTRL, HDO);
    if (waitHDOoff(hcsp) == FALSE)	/* Wait HDO off   */
	return (FALSE);

    ORC_WR(hcsp->HCS_Base + ORC_HDATA, address);	/* Write address */
    ORC_WR(hcsp->HCS_Base + ORC_HCTRL, HDO);
    if (waitHDOoff(hcsp) == FALSE)	/* Wait HDO off   */
	return (FALSE);

    if (waitHDIset(hcsp, &bData) == FALSE)	/* Wait HDI set   */
	return (FALSE);
    *pDataIn = ORC_RD(hcsp->HCS_Base, ORC_HDATA);
    ORC_WR(hcsp->HCS_Base + ORC_HSTUS, bData);	/* Clear HDI    */

    return (TRUE);
}

/***************************************************************************/
void orc_exec_scb(ORC_HCS * hcsp, ORC_SCB * scbp)
{
    scbp->SCB_Status = SCB_POST;
    ORC_WR(hcsp->HCS_Base + ORC_PQUEUE, scbp->SCB_ScbIdx);
    return;
}


/***********************************************************************

 Read SCSI H/A configuration parameters from serial EEPROM
************************************************************************/
int se2_rd_all(ORC_HCS * hcsp)
{
    int eep_addr;
    UCHAR *np, chksum = 0;

    np = (UCHAR *) nvramp;
    for (eep_addr = 0; eep_addr < NVRAM_SIZE; eep_addr++, np++) { /* <01> */
	if (get_NVRAM(hcsp, (unsigned char) eep_addr, np) == FALSE)
	    return -1;
//      *np++ = get_NVRAM(hcsp, (unsigned char ) eep_addr);
    }

/*------ Is ckecksum ok ? ------*/
    np = (UCHAR *) nvramp + 0xC0;
    for (eep_addr = 0xC0; eep_addr < NVRAM_SIZE-1; eep_addr++) {
	chksum += *np++;
    }

    if (nvramp->CheckSum != (UCHAR) chksum) {
	printk("cb32_cb : CheckSumErr %02X : %02X\n", nvramp->CheckSum, chksum);
	return -1;
    }
    return 1;
}

/************************************************************************
 Function name  : se2_update_all
 Description	: Update SCSI H/A configuration parameters
		  from serial EEPROM
*************************************************************************/
void se2_update_all(ORC_HCS * hcsp)
{				/* setup default pattern  */
    int eep_addr;
    UCHAR *np, *np1, chksum = 0;

    /* Calculate checksum first   */
    np = (UCHAR *) dftNvRam + 0xC0;
    for (eep_addr = 0xC0; eep_addr < NVRAM_SIZE-1; eep_addr++)
	chksum += *np++;
    *np = chksum;

    np = (UCHAR *) dftNvRam;
    np1 = (UCHAR *) nvramp;
    for (eep_addr = 0; eep_addr < NVRAM_SIZE; eep_addr++, np++, np1++) {
	if (*np != *np1) {
		set_NVRAM(hcsp, (unsigned char) eep_addr, *np);
	}
    }
    return;
}

/*************************************************************************
 Function name  : read_eeprom
**************************************************************************/
void read_eeprom(ORC_HCS * hcsp)
{
    if (se2_rd_all(hcsp) != 1) {
	se2_update_all(hcsp);	/* setup default pattern        */
	se2_rd_all(hcsp);	/* load again                   */
    }
}


/**************************************************************************
 Function name  : load_FW
**************************************************************************/
UCHAR load_FW(ORC_HCS * hcsp)
{
    U32 dData;
    USHORT wBIOSAddress;
    USHORT i;
    UCHAR *pData, bData;


    bData = ORC_RD(hcsp->HCS_Base, ORC_GCFG);
	/* Enable EEPROM programming */
    ORC_WR(hcsp->HCS_Base + ORC_GCFG, bData | EEPRG);
    ORC_WR(hcsp->HCS_Base + ORC_EBIOSADR2, 0x00);
    ORC_WRSHORT(hcsp->HCS_Base + ORC_EBIOSADR0, 0x00);
    if (ORC_RD(hcsp->HCS_Base, ORC_EBIOSDATA) != 0x55) {
	/* Disable EEPROM programming */
	ORC_WR(hcsp->HCS_Base + ORC_GCFG, bData);
	return (FALSE);
    }
    ORC_WRSHORT(hcsp->HCS_Base + ORC_EBIOSADR0, 0x01);
    if (ORC_RD(hcsp->HCS_Base, ORC_EBIOSDATA) != 0xAA) {
	/* Disable EEPROM programming */
	ORC_WR(hcsp->HCS_Base + ORC_GCFG, bData);
	return (FALSE);
    }
	/* Enable SRAM programming */
    ORC_WR(hcsp->HCS_Base + ORC_RISCCTL, PRGMRST | DOWNLOAD);
    pData = (UCHAR *) & dData;
    dData = 0;		/* Initial FW address to 0 */
    ORC_WRSHORT(hcsp->HCS_Base + ORC_EBIOSADR0, 0x10);
    *pData = ORC_RD(hcsp->HCS_Base, ORC_EBIOSDATA);	/* Read from BIOS */
    ORC_WRSHORT(hcsp->HCS_Base + ORC_EBIOSADR0, 0x11);
	/* Read from BIOS */
    *(pData + 1) = ORC_RD(hcsp->HCS_Base, ORC_EBIOSDATA);
    ORC_WRSHORT(hcsp->HCS_Base + ORC_EBIOSADR0, 0x12);
	/* Read from BIOS */
    *(pData + 2) = ORC_RD(hcsp->HCS_Base, ORC_EBIOSDATA);
    ORC_WR(hcsp->HCS_Base + ORC_EBIOSADR2, *(pData + 2));
    ORC_WRLONG(hcsp->HCS_Base + ORC_FWBASEADR, dData);	/* Write FW address */

    wBIOSAddress = (USHORT) dData;	/* FW code locate at BIOS address + ? */
    for (i = 0, pData = (UCHAR *) & dData;	/* Download the code    */
	    i < 0x1000;	/* Firmware code size = 4K      */
	    i++, wBIOSAddress++) {
	ORC_WRSHORT(hcsp->HCS_Base + ORC_EBIOSADR0, wBIOSAddress);
	/* Read from BIOS */
	*pData++ = ORC_RD(hcsp->HCS_Base, ORC_EBIOSDATA);
	if ((i % 4) == 3) {
	    /* Write every 4 bytes */
	    ORC_WRLONG(hcsp->HCS_Base + ORC_RISCRAM, dData);
	    pData = (UCHAR *) & dData;
	}
    }

	/* Reset program count 0 */
    ORC_WR(hcsp->HCS_Base + ORC_RISCCTL, PRGMRST | DOWNLOAD);
    wBIOSAddress -= 0x1000;	/* Reset the BIOS adddress      */
    for (i = 0, pData = (UCHAR *) & dData;	/* Check the code       */
	    i < 0x1000;	/* Firmware code size = 4K      */
	    i++, wBIOSAddress++) {
	ORC_WRSHORT(hcsp->HCS_Base + ORC_EBIOSADR0, wBIOSAddress);
	/* Read from BIOS */
	*pData++ = ORC_RD(hcsp->HCS_Base, ORC_EBIOSDATA);
	if ((i % 4) == 3) {
	    if (ORC_RDLONG(hcsp->HCS_Base, ORC_RISCRAM) != dData) {
		/* Reset program to 0 */
		ORC_WR(hcsp->HCS_Base + ORC_RISCCTL, PRGMRST);
		/*Disable EEPROM programming */
		ORC_WR(hcsp->HCS_Base + ORC_GCFG, bData);
		return (FALSE);
	    }
	    pData = (UCHAR *) & dData;
	}
    }
	/* Reset program to 0   */
    ORC_WR(hcsp->HCS_Base + ORC_RISCCTL, PRGMRST);
    ORC_WR(hcsp->HCS_Base + ORC_GCFG, bData);	/* Disable EEPROM programming */
    return (TRUE);
}

/***************************************************************************/
void
setup_SCBs(ORC_HCS * hcsp)
/*
 Function name  : setup_SCBs
*/
{
    ORC_SCB *pVirScb;
    int i;
    UCHAR j;
    ESCB *pVirEscb;
    PVOID pPhysEscb;
    PVOID tPhysEscb;

    j = 0;
    pVirScb = NULL;
    tPhysEscb = (PVOID) NULL;
    pPhysEscb = (PVOID) NULL;
	/* Setup SCB HCS_Base and SCB Size registers */
	/* Total number of SCBs */
    ORC_WR(hcsp->HCS_Base + ORC_SCBSIZE, orc_num_scb);
	/* SCB HCS_Base address 0      */
    ORC_WRLONG(hcsp->HCS_Base + ORC_SCBBASE0, hcsp->HCS_physScbArray);
	/* SCB HCS_Base address 1      */
    ORC_WRLONG(hcsp->HCS_Base + ORC_SCBBASE1, hcsp->HCS_physScbArray);

	/* setup scatter list address with one buffer */
    pVirScb = (ORC_SCB *) hcsp->HCS_virScbArray;
    pVirEscb = (ESCB *) hcsp->HCS_virEscbArray;

    for (i = 0; i < orc_num_scb; i++) {
#if LINUX_VERSION_CODE >= CVT_LINUX_VERSION(1,3,0)
	pPhysEscb = (PVOID) ((ULONG) hcsp->HCS_virEscbArray + (sizeof(ESCB) * i));
	pVirScb->SCB_SGPAddr = (U32) VIRT_TO_BUS(pPhysEscb);
	pVirScb->SCB_SensePAddr = (U32) VIRT_TO_BUS(pPhysEscb);
#else
	pPhysEscb = (PVOID) (hcsp->HCS_physEscbArray + (sizeof(ESCB) * i));
	pVirScb->SCB_SGPAddr = (U32) pPhysEscb;
	pVirScb->SCB_SensePAddr = (U32) pPhysEscb;
#endif
	pVirScb->SCB_EScb = pVirEscb;
	pVirScb->SCB_ScbIdx = i;
	pVirScb++;
	pVirEscb++;
    }

    return;
}

/***************************************************************************/
static void initAFlag(ORC_HCS * hcsp)
{
    UCHAR i, j;

    for (i = 0; i < MAX_CHANNELS; i++) {
	for (j = 0; j < 8; j++) {
		hcsp->BitAllocFlag[i][j] = 0xffffffff;
	}
    }
}

/***************************************************************************/
int init_orchid(ORC_HCS * hcsp)
{
    UBYTE *readBytep;
    USHORT revision;
    UCHAR i;

    initAFlag(hcsp);
    ORC_WR(hcsp->HCS_Base + ORC_GIMSK, 0xFF);	/* Disable all interrupt */
    if (ORC_RD(hcsp->HCS_Base, ORC_HSTUS) & RREADY) { /* Orchid is ready */
	revision = get_FW_version(hcsp);
	if (revision == 0xFFFF) {
		/* Reset Host Adapter   */
  	    ORC_WR(hcsp->HCS_Base + ORC_HCTRL, DEVRST);
	    if (waitChipReady(hcsp) == FALSE)
		return (-1);
	    cb32_load_FW(hcsp);	/* Download FW for CB32   */
		/* Setup SCB HCS_Base and SCB Size registers */
	    setup_SCBs(hcsp);
	    ORC_WR(hcsp->HCS_Base + ORC_HCTRL, 0); /* clear HOSTSTOP */
	    if (waitFWReady(hcsp) == FALSE) /* Wait for firmware ready */
		return (-1);
	    revision = get_FW_version(hcsp);
	} else {
		/* Setup SCB HCS_Base and SCB Size registers */
	    setup_SCBs(hcsp);
	}
    } else {		/* Orchid is not Ready          */
	// printk("cb32_cb : Orchid is not Ready\n");
	ORC_WR(hcsp->HCS_Base + ORC_HCTRL, DEVRST); /* Reset Host Adapter */
	if (waitChipReady(hcsp) == FALSE)
	    return (-1);
	cb32_load_FW(hcsp);	/* Download FW for CB32   */
	/* Setup SCB HCS_Base and SCB Size registers */
	setup_SCBs(hcsp);
	ORC_WR(hcsp->HCS_Base + ORC_HCTRL, HDO); /* Do Hardware Reset &  */

	/*     clear HOSTSTOP  */
	if (waitFWReady(hcsp) == FALSE) /* Wait for firmware ready      */
	    return (-1);
	revision = get_FW_version(hcsp);
    }

/*------------- get serial EEProm settting -------*/

    read_eeprom(hcsp);

#ifdef DBG
	printk("cb32_cb : Firmware Revision = %04X Nvram Revision = %d\n",
			 revision, nvramp->Revision);
#endif
    if (nvramp->Revision != 1)
	return (-1);

    hcsp->HCS_SCSI_ID = nvramp->SCSI0Id;
    hcsp->HCS_BIOS = nvramp->BIOSConfig1;
    hcsp->HCS_MaxTar = MAX_TARGETS;
    readBytep = (UCHAR *) & (nvramp->Target00Config);
    for (i = 0; i < 16; readBytep++, i++) {
	hcsp->TargetFlag[i] = *readBytep;
	hcsp->MaximumTags[i] = orc_num_scb;
    }		/* for                          */

	/* Reset SCSI bus               */
    if (nvramp->SCSI0Config & NCC_BUSRESET) {
	hcsp->HCS_Flags |= HCF_SCSI_RESET;
    }

	/* enable RP FIFO interrupt     */
    ORC_WR(hcsp->HCS_Base + ORC_GIMSK, 0xFB);
#ifdef DBG
	printk("cb32_cb : Bus Reset\n");
#endif
	orc_reset_scsi_bus(hcsp);
    return (0);
}

/*****************************************************************************
 Function name  : orc_reset_scsi_bus
 Description    : Reset registers, reset a hanging bus and
                  kill active and disconnected commands for target w/o soft reset
 Input          : pHCB  -       Pointer to host adapter structure
 Output         : None.
 Return         : pSRB  -       Pointer to SCSI request block.
*****************************************************************************/
int orc_reset_scsi_bus(ORC_HCS * pHCB)
{				/* I need Host Control Block Information */
    ULONG flags;

#ifdef DBG
	printk("cb32_cb: enter cb32_bus_reset\n");
#endif

#if LINUX_VERSION_CODE < CVT_LINUX_VERSION(2,1,95)
	save_flags(flags);
	cli();
#else
	spin_lock_irqsave(&(pHCB->BitAllocFlagLock), flags);
#endif

	initAFlag(pHCB);
	/* reset scsi bus */
	ORC_WR(pHCB->HCS_Base + ORC_HCTRL, SCSIRST);
	if (waitSCSIRSTdone(pHCB) == FALSE) {
#if LINUX_VERSION_CODE < CVT_LINUX_VERSION(2,1,95)
		restore_flags(flags);
#else
		spin_unlock_irqrestore(&(pHCB->BitAllocFlagLock), flags);
#endif
		return (SCSI_RESET_ERROR);
	} else {
#if LINUX_VERSION_CODE < CVT_LINUX_VERSION(2,1,95)
		restore_flags(flags);
#else
		spin_unlock_irqrestore(&(pHCB->BitAllocFlagLock), flags);
#endif
		return (SCSI_RESET_SUCCESS);
	}
}

/*****************************************************************************
 Function name  : orc_device_reset
 Description    : Reset registers, reset a hanging bus and
                  kill active and disconnected commands for target w/o soft reset
 Input          : pHCB  -       Pointer to host adapter structure
 Output         : None.
 Return         : pSRB  -       Pointer to SCSI request block.
*****************************************************************************/
int orc_device_reset(ORC_HCS * pHCB, ULONG SCpnt, unsigned int target, unsigned int ResetFlags)
{				/* I need Host Control Block Information */
    ORC_SCB *pScb;
    ESCB *pVirEscb;
    ORC_SCB *pVirScb;
    UCHAR i;
    ULONG flags;

#ifdef DBG
	printk("cb32_cb: enter cb32_dev_reset\n");
#endif

#if LINUX_VERSION_CODE < CVT_LINUX_VERSION(2,1,95)
	save_flags(flags);
	cli();
#else
	spin_lock_irqsave(&(pHCB->BitAllocFlagLock), flags);
#endif
    pScb = (ORC_SCB *) NULL;
    pVirEscb = (ESCB *) NULL;

	/* setup scatter list address with one buffer */
    pVirScb = (ORC_SCB *) pHCB->HCS_virScbArray;

    initAFlag(pHCB);
	/* device reset */
    for (i = 0; i < orc_num_scb; i++) {
	pVirEscb = pVirScb->SCB_EScb;
	if ((pVirScb->SCB_Status) && (pVirEscb->SCB_Srb == (unsigned char *) SCpnt))
		break;
	pVirScb++;
    }

    if (i == orc_num_scb) {
	printk("Unable to Reset - No SCB Found\n");
#if LINUX_VERSION_CODE < CVT_LINUX_VERSION(2,1,95)
	restore_flags(flags);
#else
	spin_unlock_irqrestore(&(pHCB->BitAllocFlagLock), flags);
#endif
	return (SCSI_RESET_NOT_RUNNING);
    }
    if ((pScb = orc_alloc_scb(pHCB)) == NULL) {
#if LINUX_VERSION_CODE < CVT_LINUX_VERSION(2,1,95)
	restore_flags(flags);
#else
	spin_unlock_irqrestore(&(pHCB->BitAllocFlagLock), flags);
#endif
	return (SCSI_RESET_NOT_RUNNING);
    }
    pScb->SCB_Opcode = ORC_BUSDEVRST;
    pScb->SCB_Target = target;
    pScb->SCB_HaStat = 0;
    pScb->SCB_TaStat = 0;
    pScb->SCB_Status = 0x0;
    pScb->SCB_Link = 0xFF;
    pScb->SCB_Reserved0 = 0;
    pScb->SCB_Reserved1 = 0;
    pScb->SCB_XferLen = 0;
    pScb->SCB_SGLen = 0;

    pVirEscb->SCB_Srb = 0;
    if (ResetFlags & SCSI_RESET_SYNCHRONOUS) {
	pVirEscb->SCB_Srb = (unsigned char *) SCpnt;
    }
    orc_exec_scb(pHCB, pScb);	/* Start execute SCB            */
#if LINUX_VERSION_CODE < CVT_LINUX_VERSION(2,1,95)
    restore_flags(flags);
#else
    spin_unlock_irqrestore(&(pHCB->BitAllocFlagLock), flags);
#endif
    return SCSI_RESET_PENDING;
}


/***************************************************************************
 Function name	: orc_alloc_scb
*****************************************************************************/
ORC_SCB *orc_alloc_scb(ORC_HCS * hcsp)
{
    ORC_SCB *pTmpScb;
    UCHAR Ch;
    ULONG idx;
    UCHAR index;
    UCHAR i;
    ULONG flags;

#if LINUX_VERSION_CODE < CVT_LINUX_VERSION(2,1,95)
    save_flags(flags);
    cli();
#else
    spin_lock_irqsave(&(hcsp->BitAllocFlagLock), flags);
#endif
    Ch = hcsp->HCS_Index;
    for (i = 0; i < 8; i++) {
	for (index = 0; index < 32; index++) {
	    if ((hcsp->BitAllocFlag[Ch][i] >> index) & 0x01) {
		hcsp->BitAllocFlag[Ch][i] &= ~(1 << index);
		break;
	    } // if (hcsp->...
	} // for (index = ...
	idx = index + 32 * i;
	pTmpScb = (PVOID) ((ULONG) hcsp->HCS_virScbArray + (idx * sizeof(ORC_SCB)));
#if LINUX_VERSION_CODE < CVT_LINUX_VERSION(2,1,95)
	restore_flags(flags);
#else
	spin_unlock_irqrestore(&(hcsp->BitAllocFlagLock), flags);
#endif
	return (pTmpScb);
    } // for (i = ...
#if LINUX_VERSION_CODE < CVT_LINUX_VERSION(2,1,95)
    restore_flags(flags);
#else
    spin_unlock_irqrestore(&(hcsp->BitAllocFlagLock), flags);
#endif
    return (NULL);
}


/***************************************************************************
 Function name	: orc_release_scb
*****************************************************************************/
void orc_release_scb(ORC_HCS * hcsp, ORC_SCB * scbp)
{
	ULONG flags;
	UCHAR Index;
	UCHAR i;
	UCHAR Ch;

#if LINUX_VERSION_CODE < CVT_LINUX_VERSION(2,1,95)
	save_flags(flags);
	cli();
#else
	spin_lock_irqsave(&(hcsp->BitAllocFlagLock), flags);
#endif
	Ch = hcsp->HCS_Index;
	Index = scbp->SCB_ScbIdx;
	i = Index / 32;
	Index %= 32;
	hcsp->BitAllocFlag[Ch][i] |= (1 << Index);
#if LINUX_VERSION_CODE < CVT_LINUX_VERSION(2,1,95)
	restore_flags(flags);
#else
	spin_unlock_irqrestore(&(hcsp->BitAllocFlagLock), flags);
#endif
}


/*****************************************************************************
 Function name	: Addcb32_into_Adapter_table
 Description	: This function will scan PCI bus to get all Orchid card
 Input		: None.
 Output		: None.
 Return		: SUCCESSFUL	- Successful scan
 ohterwise	- No drives founded
*****************************************************************************/
int Addcb32_into_Adapter_table(WORD wBIOS, WORD wBASE, BYTE bInterrupt,
				  BYTE bBus, BYTE bDevice)
{
	unsigned int i, j;

	for (i = 0; i < MAX_SUPPORTED_ADAPTERS; i++) {
#ifdef HAVE_BIOS
		if (cb32_adpt[i].ADPT_BIOS < wBIOS)
			continue;
		if (cb32_adpt[i].ADPT_BIOS == wBIOS) {
			if (cb32_adpt[i].ADPT_BASE == wBASE) {
				if (cb32_adpt[i].ADPT_Bus != 0xFF)
					return (FAILURE);
				else if (cb32_adpt[i].ADPT_BASE < wBASE)
					continue;
			}
		}
#endif
		for (j = MAX_SUPPORTED_ADAPTERS - 1; j > i; j--) {
			cb32_adpt[j].ADPT_BASE = cb32_adpt[j - 1].ADPT_BASE;
			cb32_adpt[j].ADPT_INTR = cb32_adpt[j - 1].ADPT_INTR;
			cb32_adpt[j].ADPT_BIOS = cb32_adpt[j - 1].ADPT_BIOS;
			cb32_adpt[j].ADPT_Bus = cb32_adpt[j - 1].ADPT_Bus;
			cb32_adpt[j].ADPT_Device = cb32_adpt[j - 1].ADPT_Device;
		}
		cb32_adpt[i].ADPT_BASE = wBASE;
		cb32_adpt[i].ADPT_INTR = bInterrupt;
		cb32_adpt[i].ADPT_BIOS = wBIOS;
		cb32_adpt[i].ADPT_Bus = bBus;
		cb32_adpt[i].ADPT_Device = bDevice;
		return (SUCCESSFUL);
	}
	return (FAILURE);
}


/*****************************************************************************
 Function name	: init_cb32Adapter_table
 Description	: This function will scan PCI bus to get all Orchid card
 Input		: None.
 Output		: None.
 Return		: SUCCESSFUL	- Successful scan
 ohterwise	- No drives founded
*****************************************************************************/
void init_cb32Adapter_table(void)
{
	int i;

	for (i = 0; i < MAX_SUPPORTED_ADAPTERS; i++) {	/* Initialize adapter structure */
		cb32_adpt[i].ADPT_BIOS = 0xffff;
		cb32_adpt[i].ADPT_BASE = 0xffff;
		cb32_adpt[i].ADPT_INTR = 0xff;
		cb32_adpt[i].ADPT_Bus = 0xff;
		cb32_adpt[i].ADPT_Device = 0xff;
	}
}

/*****************************************************************************
 Function name  : get_orcPCIConfig
 Description    : 
 Input          : pHCB  -       Pointer to host adapter structure
 Output         : None.
 Return         : pSRB  -       Pointer to SCSI request block.
*****************************************************************************/
void get_orcPCIConfig(ORC_HCS * pCurHcb, int ch_idx)
{
	pCurHcb->HCS_Base = cb32_adpt[ch_idx].ADPT_BASE;	/* Supply base address  */
	pCurHcb->HCS_BIOS = cb32_adpt[ch_idx].ADPT_BIOS;	/* Supply BIOS address  */
	pCurHcb->HCS_Intr = cb32_adpt[ch_idx].ADPT_INTR;	/* Supply interrupt line */
	return;
}


/*****************************************************************************
 Function name  : abort_SCB
 Description    : Abort a queued command.
	                 (commands that are on the bus can't be aborted easily)
 Input          : pHCB  -       Pointer to host adapter structure
 Output         : None.
 Return         : pSRB  -       Pointer to SCSI request block.
*****************************************************************************/
int abort_SCB(ORC_HCS * hcsp, ORC_SCB * pScb)
{
    unsigned char bData, bStatus;

#ifdef DBG
    printk("cb32_cb: abort SCB \n");
#endif
    ORC_WR(hcsp->HCS_Base + ORC_HDATA, ORC_CMD_ABORT_SCB);	/* Write command */
    ORC_WR(hcsp->HCS_Base + ORC_HCTRL, HDO);
    if (waitHDOoff(hcsp) == FALSE)	/* Wait HDO off   */
	return (FALSE);

    ORC_WR(hcsp->HCS_Base + ORC_HDATA, pScb->SCB_ScbIdx);	/* Write address */
    ORC_WR(hcsp->HCS_Base + ORC_HCTRL, HDO);
    if (waitHDOoff(hcsp) == FALSE)	/* Wait HDO off   */
	return (FALSE);

    if (waitHDIset(hcsp, &bData) == FALSE)	/* Wait HDI set   */
	return (FALSE);
    bStatus = ORC_RD(hcsp->HCS_Base, ORC_HDATA);
    ORC_WR(hcsp->HCS_Base + ORC_HSTUS, bData);	/* Clear HDI    */

    if (bStatus == 1)	/* 0 - Successfully               */
	return (FALSE);	/* 1 - Fail                     */

    return (TRUE);
}

/*****************************************************************************
 Function name  : cb32_abort
 Description    : Abort a queued command.
	                 (commands that are on the bus can't be aborted easily)
 Input          : pHCB  -       Pointer to host adapter structure
 Output         : None.
 Return         : pSRB  -       Pointer to SCSI request block.
*****************************************************************************/
int orc_abort_srb(ORC_HCS * hcsp, ULONG SCpnt)
{
    ESCB *pVirEscb;
    ORC_SCB *pVirScb;
    UCHAR i;
    ULONG flags;

#ifdef DBG
    printk("cb32_cb: abort SRB \n");
#endif
#if LINUX_VERSION_CODE < CVT_LINUX_VERSION(2,1,95)
    save_flags(flags);
    cli();
#else
    spin_lock_irqsave(&(hcsp->BitAllocFlagLock), flags);
#endif

    pVirScb = (ORC_SCB *) hcsp->HCS_virScbArray;

    for (i = 0; i < orc_num_scb; i++, pVirScb++) {
	pVirEscb = pVirScb->SCB_EScb;
	if ((pVirScb->SCB_Status) && (pVirEscb->SCB_Srb == (unsigned char *) SCpnt)) {
	    if (pVirScb->SCB_TagMsg == 0) {
#if LINUX_VERSION_CODE < CVT_LINUX_VERSION(2,1,95)
		restore_flags(flags);
#else
		spin_unlock_irqrestore(&(hcsp->BitAllocFlagLock), flags);
#endif
#ifdef DBG
    printk(" busy\n");
#endif
		return (SCSI_ABORT_BUSY);
	    } else {
		if (abort_SCB(hcsp, pVirScb)) {
		    pVirEscb->SCB_Srb = NULL;
#if LINUX_VERSION_CODE < CVT_LINUX_VERSION(2,1,95)
		    restore_flags(flags);
#else
		    spin_unlock_irqrestore(&(hcsp->BitAllocFlagLock), flags);
#endif
#ifdef DBG
    printk(" success\n");
#endif
		    return (SCSI_ABORT_SUCCESS);
		} else {
#if LINUX_VERSION_CODE < CVT_LINUX_VERSION(2,1,95)
		    restore_flags(flags);
#else
		    spin_unlock_irqrestore(&(hcsp->BitAllocFlagLock), flags);
#endif
#ifdef DBG
    printk(" not running\n");
#endif
		    return (SCSI_ABORT_NOT_RUNNING);
		} // if (abort_SCB()
	    } //  if (pVirScb->SCB_TagMsg == ...
	}  // if (pVirScb->SCB_Status ...
    } // for ( i = ...
#if LINUX_VERSION_CODE < CVT_LINUX_VERSION(2,1,95)
    restore_flags(flags);
#else
    spin_unlock_irqrestore(&(hcsp->BitAllocFlagLock), flags);
#endif
#ifdef DBG
    printk(" not running\n");
#endif
    return (SCSI_ABORT_NOT_RUNNING);
}

/***********************************************************************
 Routine Description:
	  This is the interrupt service routine for the Orchid SCSI adapter.
	  It reads the interrupt register to determine if the adapter is indeed
	  the source of the interrupt and clears the interrupt at the device.
 Arguments:
	  HwDeviceExtension - HBA miniport driver's adapter data storage
 Return Value:
***********************************************************************/
void
orc_interrupt(ORC_HCS * hcsp)
{
    BYTE bQueCnt;
    BYTE bScbIdx;
    ORC_SCB *pScb;


    if ((bQueCnt = ORC_RD(hcsp->HCS_Base, ORC_RQUEUECNT)) == 0) {
	return;		// (FALSE);
    }
    if (bQueCnt == 0xff) {	// Unpluged Card
#ifdef DBG
	printk("cb32_cb: Unpluged Card\n");
#endif
	return;		// (FALSE);
    }

    do {
	bScbIdx = ORC_RD(hcsp->HCS_Base, ORC_RQUEUE);

	pScb = (ORC_SCB *) ((ULONG) hcsp->HCS_virScbArray + (ULONG) (sizeof(ORC_SCB) * bScbIdx));
	pScb->SCB_Status = 0x0;

	cb32SCBPost((BYTE *) hcsp, (BYTE *) pScb);
    } while (ORC_RD(hcsp->HCS_Base, ORC_RQUEUECNT));
    return;			//(TRUE);

}				/* End of I850Interrupt() */

// #include "load_fw.c"

